// Crypto.cpp: implementation of the CCrypto class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Crypto.h"
#include <wincrypt.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CCrypto::CCrypto()
{
	byPublicKey = NULL;
	dwPublicKeyLength = 0;

	byPrivateKey = NULL;
	dwPrivateKeyLength = 0;
	
	bySession = NULL;
	dwSessionLength = 0;

	bySignature = NULL;
	dwSignatureLength = 0;

	hCryptProv  = NULL;
	hKeyPair    = NULL;
	hSessionKey = NULL;

	lType = 0;
}

CCrypto::~CCrypto()
{
	//Clear up any allocated buffers...
	if (byPublicKey != NULL)
		delete byPublicKey;

	if (byPrivateKey != NULL)
		delete byPrivateKey;

	if (bySession != NULL)
		delete bySession;

	if (bySignature != NULL)
		delete bySignature;
}

boolean CCrypto::SessionStart()
{
	LPCTSTR KEY_CONTAINER    = NULL;
	LPCTSTR SERVICE_PROVIDER = L"Microsoft Base Cryptographic Provider v1.0";
									    
    if (!CryptAcquireContext(&hCryptProv, KEY_CONTAINER, SERVICE_PROVIDER, PROV_RSA_FULL, 0))
	{		
	    if  (!CryptAcquireContext(&hCryptProv, KEY_CONTAINER, SERVICE_PROVIDER, PROV_RSA_FULL, CRYPT_NEWKEYSET))
		{			
			return FALSE;
		}
	}
	
	return TRUE;
}

void CCrypto::SessionEnd()
{
    //Release any session key handle...
    if (hSessionKey != 0)
		CryptDestroyKey(hSessionKey);

	//Release any key pair handle...
    if (hKeyPair != 0) 
		CryptDestroyKey(hKeyPair);

    //Release provider handle...
    if (hCryptProv != 0)
		CryptReleaseContext(hCryptProv, 0);
}

/*boolean CCrypto::GenerateKeyPair(boolean bUseExchangeKeyPair)
{
    long lParams;

    //Set the key length to 512, allow the keys to be exported, no SALT...
    lParams = 0x2000000 | CRYPT_EXPORTABLE | CRYPT_NO_SALT;

    //Set the type of key pair Exchange or Signature...
    if (bUseExchangeKeyPair)
        lType = AT_KEYEXCHANGE;
    else
        lType = AT_SIGNATURE;
    
    //Release old key pair handle...
    if (hKeyPair != NULL)
		CryptDestroyKey(hKeyPair);

    //Generate the key pair...
    if (!CryptGenKey(hCryptProv, lType, lParams, &hKeyPair))    
        return FALSE;

	return TRUE;
}*/

boolean CCrypto::ExportKeyPair(const char* strPassword)
{

    //Export the public key to strPublicBlob (Key !encrypted)...
    if (!CryptExportKey(hKeyPair, 0, PUBLICKEYBLOB, 0, NULL, &dwPublicKeyLength))
		return FALSE;
    
	//Do we have an old PublicKey???
	if (byPublicKey != NULL)
		delete [] byPublicKey;

	byPublicKey = new BYTE[dwPublicKeyLength];

    if (!CryptExportKey(hKeyPair, 0, PUBLICKEYBLOB, 0, byPublicKey, &dwPublicKeyLength))    
        return FALSE;    

    //Export the private key which is encrypted with a password...
    //Need session key derived from a password...
    SessionKeyFromPassword ((BYTE *)strPassword, (DWORD)strlen(strPassword));

    if (!CryptExportKey(hKeyPair, hSessionKey, PRIVATEKEYBLOB, 0, NULL, &dwPrivateKeyLength))	
		return FALSE;
    
	//Do we have an old private key???
	if (byPrivateKey != NULL)
		delete [] byPrivateKey;

	byPrivateKey = new BYTE[dwPrivateKeyLength];

    if (!CryptExportKey(hKeyPair, hSessionKey, PRIVATEKEYBLOB, 0, byPrivateKey, &dwPrivateKeyLength))
		return FALSE;

	if (hSessionKey != 0)
	{
		CryptDestroyKey(hSessionKey);
		hSessionKey = 0;
	}

	return TRUE;
}

boolean CCrypto::SessionKeyFromPassword(BYTE* byKey, DWORD dwLength)
{
    long   lParams;
    BYTE*  byHash;
    DWORD  dwHashLen;
    HCRYPTHASH hHash;

    //Create a hash object to calculate a session key from the Password...
    //Could use CALG_MD5 to get a 128-bit hash.  CALG_SHA returns a 160-bit hash...
    if (!CryptCreateHash(hCryptProv, CALG_SHA, 0, 0, &hHash))
        return FALSE;    
    
    //Hash the Password...
    if (!CryptHashData(hHash, byKey, dwLength, 0))    
        return FALSE;    

    //Get the actual hash value first the length...
    CryptGetHashParam(hHash, HP_HASHVAL, NULL, &dwHashLen, 0);
    byHash = new BYTE[dwHashLen];

	//Now the actual value...
    if (!CryptGetHashParam(hHash, HP_HASHVAL, byHash, &dwHashLen, 0))
		return FALSE;
	
	if (byHash != NULL)
		delete [] byHash;

    //Set the key to be exported, no SALT...
    lParams = CRYPT_EXPORTABLE;
    
    //Release old session key handle if one exists...
    if (hSessionKey != NULL) 
		CryptDestroyKey(hSessionKey);

    //Derive a session key from the hash object...
    if (!CryptDeriveKey(hCryptProv, CALG_RC4, hHash, lParams, &hSessionKey))
		return FALSE;
    
    //Destroy the hash object...
    if (hHash != NULL)
		CryptDestroyHash(hHash);

	return TRUE;
}

boolean CCrypto::ImportKeyPair(const char* strPassword, boolean bExchangeKeyPair)
{
    long lParams;

    //Check to see if there are no keys available at all...
    if ((byPrivateKey == NULL) && (byPublicKey == NULL))
		return FALSE;

    //Release old key pair handle...
    if (hKeyPair == NULL)
        CryptDestroyKey(hKeyPair);

    //Set the type of key pair Exchange or Signature...
    if (bExchangeKeyPair)
        lType = AT_KEYEXCHANGE;
    else
        lType = AT_SIGNATURE;

    //Public Key...
    if (byPrivateKey == NULL)
	{
        //Import the key...        
        if (!CryptImportKey(hCryptProv, byPublicKey, dwPublicKeyLength, 0, 0, &hKeyPair))
			return FALSE;
	}
    else
	{
        //Private Key...
        //Need session key derived from a password...
        SessionKeyFromPassword((BYTE*)strPassword, (DWORD)strlen(strPassword));
    
        //Import the key...        
        lParams = CRYPT_EXPORTABLE | CRYPT_NO_SALT;
        if (!CryptImportKey(hCryptProv, byPrivateKey, dwPrivateKeyLength, hSessionKey, lParams, &hKeyPair))
			return FALSE;
    }

    //Release session key handle...
    if (hSessionKey != NULL)
	{
        CryptDestroyKey(hSessionKey);
		hSessionKey = 0;
	}
    
	return TRUE;
}

BYTE* CCrypto::RandomKey(DWORD dwLength)
{
    BYTE* byKey;

	byKey = new BYTE[dwLength];
    
    CryptGenRandom(hCryptProv, dwLength, byKey);
        
    return byKey;
}

boolean CCrypto::EncryptDecrypt(BYTE *byBuffer, DWORD dwBufferLen, BYTE *byKey, DWORD dwKeyLen, boolean bEncrypt)
{
    //Get a session key derived from the password...
    SessionKeyFromPassword(byKey, dwKeyLen);

    //Encrypt or decrypt depending on the Encrypt parameter...    
    if (bEncrypt)
	{
        if (!CryptEncrypt(hSessionKey, 0, 1, 0, byBuffer, &dwBufferLen, dwBufferLen))
			return FALSE;
	}
    else
	{
        if (!CryptDecrypt(hSessionKey, 0, 1, 0, byBuffer, &dwBufferLen))
			return FALSE;
    }        

    //Destroy the session key...
    if (hSessionKey != NULL)
	{
		CryptDestroyKey(hSessionKey);
		hSessionKey = 0;
	}

	return TRUE;
}

boolean CCrypto::EncryptDecryptKeyPair(BYTE *byBuffer, DWORD dwBufferLen, boolean bEncrypt)
{
    long lParams;

    //Set the key length to 512, allow the keys to be exported, no SALT...
    lParams = CRYPT_EXPORTABLE ;

    //Encrypt using the Public Key...
    if (bEncrypt)
	{
	    //Get a random session key...
		if (!CryptGenKey(hCryptProv, CALG_RC4, lParams, &hSessionKey))
			return FALSE;

        //Export and encrypt the session key...
        if (!CryptExportKey(hSessionKey, hKeyPair, SIMPLEBLOB, 0, NULL, &dwSessionLength))
		{
			long lErr = GetLastError();
			return FALSE;
		}

		if (bySession != NULL)
			delete [] bySession;

        bySession = new BYTE[dwSessionLength];

        if (!CryptExportKey(hSessionKey, hKeyPair, SIMPLEBLOB, 0, bySession, &dwSessionLength))
            return FALSE;

        //Encrypt the data        
        if (!CryptEncrypt(hSessionKey, 0, 1, 0, byBuffer, &dwBufferLen, dwBufferLen))
			return FALSE;
	}
	else
	{    
        //Import the session key...        
        if (!CryptImportKey(hCryptProv, bySession, dwSessionLength, 0, 0, &hSessionKey))
			return FALSE;
        
        //Decrypt...        
        if (!CryptDecrypt(hSessionKey, 0, 1, 0, byBuffer, &dwBufferLen))
			return FALSE;
	}

    //Destroy the session key...
    if (hSessionKey != NULL)
	{
		CryptDestroyKey(hSessionKey);
		hSessionKey = 0;
	}

	return TRUE;
}

BYTE* CCrypto::Hash(BYTE *byBuffer, DWORD dwBufferLen, DWORD* dwHashLen, ALG_ID Algid)
{    
	HCRYPTHASH hHash;
	BYTE*      byHash;

    //Create a hash object...
    if (!CryptCreateHash(hCryptProv, Algid, 0, 0, &hHash))
		return FALSE;
    
    //Hash the buffer...
    if (!CryptHashData(hHash, byBuffer, dwBufferLen, 0))
		return FALSE;
    
    //Get the actual hash value...
    if (!CryptGetHashParam(hHash, HP_HASHVAL, NULL, dwHashLen, 0))
		return FALSE;

    byHash = new BYTE[*dwHashLen];
	
    if (!CryptGetHashParam(hHash, HP_HASHVAL, byHash, dwHashLen, 0))
		return FALSE;
    
    //Destroy the hash object...
    if (hHash != NULL)
		CryptDestroyHash(hHash);

	return byHash;
}

boolean CCrypto::SignValidate(BYTE *byBuffer, DWORD dwBufferLen, boolean bSign)
{
    HCRYPTHASH hHash;
    
    //Create a hash object to hash the input data...
    if (!CryptCreateHash(hCryptProv, CALG_SHA, 0, 0, &hHash))
		return FALSE;
    
    //Hash the data
    if (!CryptHashData(hHash, byBuffer, dwBufferLen, 0))
	{
		CryptDestroyHash(hHash);
		return FALSE;
	}
    
    if (bSign)
	{
        //Determine the size of signature...
        if (!CryptSignHash(hHash, AT_SIGNATURE, 0, 0, NULL, &dwSignatureLength))
		{
			CryptDestroyHash(hHash);
			return FALSE;
		}
		
		if (bySignature != NULL)
			delete [] bySignature;

		bySignature = new BYTE[dwSignatureLength];

        //Sign hash object...
        if (!CryptSignHash(hHash, AT_SIGNATURE, 0, 0, bySignature, &dwSignatureLength))
		{
			CryptDestroyHash(hHash);
			return FALSE;
		}

		CryptDestroyHash(hHash);

        return TRUE;                
	}
    else
	{
        //Validate the signature...        
        if (!CryptVerifySignature(hHash, bySignature, dwSignatureLength, hKeyPair, 0, 0))
		{
			CryptDestroyHash(hHash);
			return FALSE;
		}
				
    }

	CryptDestroyHash(hHash);
	
	return TRUE;
}

BOOL CCrypto::SetPublicKey(BYTE *PublicKey, DWORD PublicKeyLen)
{
	if (byPublicKey != NULL)
		delete byPublicKey;

	byPublicKey = new BYTE[PublicKeyLen];

	memcpy(byPublicKey, PublicKey, PublicKeyLen);

	return TRUE;
}
