#include "stdafx.h"

#include "HttpSocket.h"
#include "Logging.h"

#define AgentName L"PINsafe SAS Agent Requestor 2.0"
LPCWSTR POST_CONTENT_TYPE = L"Content-Type: text/xml";
LPCWSTR POST_ACCEPT_TYPE = L"*/*";

HttpSocket::HttpSocket(void)
{
	Headers = NULL;
	hIS = NULL;
	hCO = NULL;
	hIO = WinHttpOpen(AgentName, WINHTTP_ACCESS_TYPE_NO_PROXY, WINHTTP_NO_PROXY_NAME, WINHTTP_NO_PROXY_BYPASS, 0);
	cDataSize = 0;
	secure = FALSE;
}

HttpSocket::~HttpSocket(void)
{
	if (hIO != NULL) {
		WinHttpCloseHandle(hIO);
	}
}

//get a page, if post is false, HTTP GET is used othervise HTTP POST is used. if PostDataLength is -1 the data must be NULL terminated...	
LPSTR HttpSocket::GetPageBytes(LPCWSTR url, BOOL Post, LPCWSTR PostData, int PostDataLength, int *byteCount) {
	//did we get a url?
	if (url == NULL)
	{
		return NULL;
	}
	char *szPostData = NULL;

	//get the page and store it in ReceivedData...
	if (Post)
	{
		//Need to convert PostData from UNICODE to char*
		if (PostData != NULL) {
			int a = (int)wcslen(PostData)+1;
			szPostData = new char[a];
			WideCharToMultiByte(CP_ACP, 
								0, 
								PostData, 
								-1, 
								szPostData, 
								a,
								NULL,
								NULL);
		}

		//use http post...
		if (!PostUrl(url, szPostData, PostDataLength)) {
			//Free the PostData...
			if (szPostData) {
				delete[] szPostData;
			}
			return NULL;
		}
	}
	else
	{
		//use http get
		if (!GetFromUrl(url)) return NULL;
	}

	//Always get the headers when we get a page...
	GetHeaders();

	const int rdsize = 0x2000;
	char mr[rdsize];
	DWORD rd;
	int curpos = 0;
	cDataSize = 0;
	int rdcount = rdsize + 1;
	LPSTR response = (LPSTR)GlobalAlloc(GPTR, rdcount);

	while (WinHttpReadData(hIS, mr, rdsize - 1, &rd))
	{
		if (rd == 0) break;
		mr[rd] = '\0';
		if (curpos + rd > rdcount) {
			rdcount = curpos + rdsize + 1;
			response = (LPSTR)GlobalReAlloc(response, rdcount, GPTR);
		}
		memcpy(response+curpos, mr, rd);
		curpos += rd;
		response[curpos] = '\0';
		cDataSize =  curpos+rdsize;
	}
	if (byteCount != NULL) {
		*byteCount = cDataSize;
	}

	//Free the PostData...
	if (szPostData) {
		delete[] szPostData;
	}
	return response;
}

LPWSTR HttpSocket::GetPage(LPCWSTR url, BOOL Post, LPCWSTR PostData, int PostDataLength) {
	LPSTR request = GetPageBytes(url, Post, PostData, PostDataLength);
	if (request == NULL) {
		return NULL;
	}
	int bufLen = strlen(request);
	LPWSTR ReceivedData = new wchar_t[bufLen+1];
	MultiByteToWideChar(CP_ACP, 0, request, bufLen, ReceivedData, bufLen+1);
	GlobalFree(request);

	return ReceivedData;
}

//open a page using http post
BOOL HttpSocket::PostUrl(LPCWSTR url, LPCSTR PostData, int PostDataLength) {
	return OpenUrl(url, L"POST", PostData, PostDataLength);
}

BOOL HttpSocket::GetFromUrl(LPCWSTR url) {
	return OpenUrl(url, L"GET", NULL, 0);
}

BOOL HttpSocket::OpenUrl(LPCWSTR url, LPCWSTR verb, LPCSTR PostData, int PostDataLength) {
	DWORD error;
	//check length of postdata
	if (PostDataLength == -1) {
		PostDataLength = (int)strlen(PostData);
	}

	//some variable that we need...
	URL_COMPONENTS uc;
	//let's split the url...
	uc.dwStructSize = sizeof(uc);
	uc.lpszScheme = NULL;
	uc.dwSchemeLength = 0;
	uc.lpszHostName = NULL;
	uc.dwHostNameLength = 1;
	uc.nPort = 0;
	uc.lpszUserName = NULL;
	uc.dwUserNameLength = 0;
	uc.lpszPassword = NULL;
	uc.dwPasswordLength = 0;
	uc.lpszUrlPath = NULL;
	uc.dwUrlPathLength = 1;
	uc.lpszExtraInfo = NULL;
	uc.dwExtraInfoLength = 0;

	if (!WinHttpCrackUrl(url, 0, 0, &uc)) {
		error = GetLastError();
		LogError(L"Error in WinHttpCrackUrl: url = %s, error = %lx", url, error);
		return false;
	}

	//post the data...
	if (hCO != NULL) {
		WinHttpCloseHandle(hCO);
		hCO = NULL;
	}
	
	LPWSTR HostName = _wcsdup(uc.lpszHostName);
	HostName[uc.dwHostNameLength] = '\0';
//	delete uc.lpszHostName;

	LPWSTR FileName = _wcsdup(uc.lpszUrlPath);
	FileName[uc.dwUrlPathLength] = '\0';
//	delete uc.lpszUrlPath;

	if (hIS != NULL) {
		WinHttpCloseHandle(hIS); //if open, close the handle to the connection
		hIS = NULL;
	}
	
	DWORD flags;

	flags = WINHTTP_FLAG_REFRESH;
	if (secure) {
		flags |= WINHTTP_FLAG_SECURE;
	}

	LPCWSTR headers = POST_CONTENT_TYPE; //content type for post...
	LPCWSTR szAccept = POST_ACCEPT_TYPE; //we accept everything...
	LPCWSTR AcceptTypes[2]={0};
	AcceptTypes[0] = szAccept;

	hCO = WinHttpConnect(hIO, HostName, uc.nPort, 0);

	BOOL result = false;
	if (hCO) {
		hIS = WinHttpOpenRequest(hCO, verb, FileName, NULL, WINHTTP_NO_REFERER, AcceptTypes, flags);
		if (hIS) {
			if (allowSelfCert) {
				LONG secFlags = SECURITY_FLAG_IGNORE_CERT_CN_INVALID | SECURITY_FLAG_IGNORE_CERT_DATE_INVALID | SECURITY_FLAG_IGNORE_UNKNOWN_CA;
				WinHttpSetOption(hIS, WINHTTP_OPTION_SECURITY_FLAGS, &secFlags, sizeof(LONG));
			}

			if (WinHttpSendRequest(hIS, headers, (DWORD)wcslen(headers), (LPVOID)PostData, PostDataLength, PostDataLength, NULL)) {
				if (WinHttpReceiveResponse(hIS, NULL)) {
					result = true;
				} else {
					error = GetLastError();
					LogError(L"Error in WinHttpReceiveResponse: %lx", error);
				}
			} else {
				error = GetLastError();
				LogError(L"Error in WinHttpSendRequest: %lx", error);
			}
		} else {
			error = GetLastError();
			LogError(L"Error in WinHttpOpenRequest: %lx", error);
		}
	} else {
		error = GetLastError();
		LogError(L"Error in WinHttpConnect: %lx", error);
	}

	free(HostName);
	free(FileName);

	return result;
}

//return a pointer to the headers from a url
LPWSTR HttpSocket::GetHeaders() {

	//Check we have something to use?
	if (hIS == NULL) {
		return NULL;
	}

	//delete old headers...
	if (Headers != NULL) {
		delete Headers;
	}

	//get the size headers
	DWORD d = 1, d2 = 0;
	int i = WinHttpQueryHeaders(hIS, WINHTTP_QUERY_RAW_HEADERS, WINHTTP_HEADER_NAME_BY_INDEX, WINHTTP_NO_OUTPUT_BUFFER, &d, &d2);
	//alloc some space for the headers
	Headers = new wchar_t[d];

	if (!WinHttpQueryHeaders(hIS, WINHTTP_QUERY_RAW_HEADERS, WINHTTP_HEADER_NAME_BY_INDEX, Headers, &d, &d2)) {
		return NULL;
	}

	return Headers;
}