#pragma once

#include "RegSettings.h"
#include "ocidl.h"

typedef enum {
	LOGIN_FAILED,
	LOGIN_PASSED,
	LOGIN_PIN_EXPIRES_SOON,
	LOGIN_PINCHANGE_REQUIRED,
	LOGIN_INCOMPLETE,
	LOGIN_UNKNOWN
} LoginState;

class PINsafeConnect
{
public:
	PINsafeConnect(void);
	~PINsafeConnect(void);
	BOOL StartSession(LPWSTR username, LPWSTR sessionId, int maxChars);
	IPicture* GetSCImageBySessionId(LPCWSTR sessionId);
	IPicture* GetSCImageByUsername(LPCWSTR username);
	void RequestSecurityString(LPCWSTR username);
	LoginState Authenticate(LPCWSTR username, LPCWSTR password, LPCWSTR otc, BOOL requestStrings=TRUE);
	BOOL ChangePIN(LPCWSTR username, LPCWSTR oldPassword, LPCWSTR newPassword, LPCWSTR oldOtc, LPCWSTR newOtc);
	void GetSettings(RegSettings* settings);
	BOOL Ping();
	BOOL LoginRemote() { return loginRemote; }
	BOOL ShowGetImage() { return showGetImage; }
	BOOL AllowDefaultLogin() { return allowDefaultLogin; }
	BOOL UseLocalTURing() { return useLocal; }
	BOOL UserExists(LPCWSTR username);

private:
	LPWSTR agentUrl;
	LPWSTR scImageUrl;
	LPWSTR dcImageUrl;
	LPWSTR secret;
	BOOL secure;
	BOOL loginRemote;
	BOOL showGetImage;
	BOOL allowDefaultLogin;
	BOOL allowSelfCert;
	BOOL useLocal;

	IPicture* GetSCImage(LPCWSTR url);
	LPWSTR CreateStartSessionRequest(LPCWSTR username);
	LPWSTR CreateLoginRequest(LPCWSTR username, LPCWSTR password, LPCWSTR otc, BOOL requestStrings);
	LPWSTR CreateChangePINRequest(LPCWSTR username, LPCWSTR oldPassword, LPCWSTR newPassword, LPCWSTR oldOtc, LPCWSTR newOtc);
	LPWSTR CreateExistsRequest(LPCWSTR username);
	BOOL TryStartSession(LPCWSTR username, LPWSTR sessionId, int maxChars);
	IStream* GetResponseStream(LPCWSTR url);

};
