//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) 2006 Microsoft Corporation. All rights reserved.
//
//

#include "stdafx.h"
#ifndef WIN32_NO_STATUS
#include <ntstatus.h>
#define WIN32_NO_STATUS
#endif
#include <lm.h>

#include "PINsafeLoginCredential.h"
#include "PINsafeLoginProvider.h"
#include "guid.h"
#include "Logging.h"
#include "TuringThread.h"
#include "CredentialCache.h"

// Global variables for handling the TURing dialog
TuringThread* _turingThread = NULL;
BOOL _dialogTerminated = FALSE;

// PINsafeLoginCredential ////////////////////////////////////////////////////////

// NOTE: Please read the readme.txt file to understand when it's appropriate to
// wrap an another credential provider and when it's not.  If you have questions
// about whether your scenario is an appropriate use of wrapping another credprov,
// please contact credprov@microsoft.com
PINsafeLoginCredential::PINsafeLoginCredential(PINsafeLoginProvider* provider):
    _cRef(1),
    _pCredProvCredentialEvents(NULL),
	_provider(provider)
{
    DllAddRef();

    SecureZeroMemory(_rgCredProvFieldDescriptors, sizeof(_rgCredProvFieldDescriptors));
    SecureZeroMemory(_rgFieldStatePairs, sizeof(_rgFieldStatePairs));
    SecureZeroMemory(_rgFieldStrings, sizeof(_rgFieldStrings));
    _pWrappedCredential = NULL;
    _dwWrappedDescriptorCount = 0;
	_credentialSelected = FALSE;
	_currentOTC = NULL;
	_username = NULL;
	_newOTC = NULL;
	_newOTC2 = NULL;
	_password = NULL;
	_newPassword = NULL;
	_newPassword2 = NULL;
	_turingLocal = NULL;
	// Get the current domain or computer name.
	LPWSTR domain ,compName;
	DWORD dsize = 0;
	LPWKSTA_INFO_100 wkInfo;
	NetWkstaGetInfo(NULL, 100, (LPBYTE*)&wkInfo);
	SHStrDup(wkInfo->wki100_langroup, &domain);
	SHStrDup(wkInfo->wki100_computername, &compName);
	_domain = new ExpandableString(domain);
	_computerName = new ExpandableString(compName);
	CoTaskMemFree(domain);
	CoTaskMemFree(compName);
	NetApiBufferFree(wkInfo);
	_settings = RegSettings::GetSettings();
	_isChangePIN = provider->IsChangePIN();
	_isCredUI = provider->IsCredUI();
	_useLocal = _settings->UseLocalAuth();
	_isShowingImage = FALSE;
	_isPINsafeAvailable = TRUE;
	_PINChangeRequired = FALSE;
	_allowUnknownUsers = _settings->AllowUnknownUsers();
	_userExists = TRUE;
	_oldUsername = NULL;
	srand(time(NULL));
}

PINsafeLoginCredential::~PINsafeLoginCredential()
{
    for (int i = 0; i < ARRAYSIZE(_rgFieldStrings); i++)
    {
        CoTaskMemFree(_rgFieldStrings[i]);
        CoTaskMemFree(_rgCredProvFieldDescriptors[i].pszLabel);
    }
	if (_turingLocal != NULL) {
		delete _turingLocal;
		_turingLocal = NULL;
	}
	_turingThread->Terminate();
	_turingThread = NULL;
	if (_currentOTC != NULL) delete _currentOTC;
	if (_username != NULL) delete _username;
	if (_newOTC != NULL) delete _newOTC;
	if (_newOTC2 != NULL) delete _newOTC2;
	if (_password != NULL) delete _password;
	if (_newPassword != NULL) delete _newPassword;
	if (_newPassword2 != NULL) delete _newPassword2;
    DllRelease();
}

int	PINsafeLoginCredential::GetFieldOffset(DWORD fieldID)
{
	return _provider->GetFieldOffset(fieldID);
}

// Initializes one credential with the field information passed in. We also keep track
// of our wrapped credential and how many fields it has.
HRESULT PINsafeLoginCredential::Initialize(
    const CREDENTIAL_PROVIDER_FIELD_DESCRIPTOR* rgcpfd,
    const FIELD_STATE_PAIR* rgfsp,
    ICredentialProviderCredential *pWrappedCredential,
    DWORD dwWrappedDescriptorCount
    )
{
    HRESULT hr = S_OK;

    // Grab the credential we're wrapping for future reference.
    if (_pWrappedCredential != NULL)
    {
        _pWrappedCredential->Release();
    }
    _pWrappedCredential = pWrappedCredential;
    _pWrappedCredential->AddRef();

    // We also need to remember how many fields the inner credential has.
    _dwWrappedDescriptorCount = dwWrappedDescriptorCount;

    // Copy the field descriptors for each field. This is useful if you want to vary the field
    // descriptors based on what Usage scenario the credential was created for.
    for (DWORD i = 0; SUCCEEDED(hr) && i < ARRAYSIZE(_rgCredProvFieldDescriptors); i++)
    {
        _rgFieldStatePairs[i] = rgfsp[i];
        hr = FieldDescriptorCopy(rgcpfd[i], &_rgCredProvFieldDescriptors[i]);
    }

	// Clear variables
	if (_currentOTC != NULL) _currentOTC->Clear();
	if (_newOTC != NULL) _newOTC->Clear();
	if (_newOTC2 != NULL) _newOTC2->Clear();
	if (_username != NULL) _username->Clear();
	if (_password != NULL) _password->Clear();

	HINSTANCE hResource = _provider->GetResourceHandle();
	int TMP_STR_LEN = 0x80;
	LPWSTR tmpStr = new wchar_t[TMP_STR_LEN];

	// Initialize the String value of all of our fields.
	if (_isChangePIN) {
		if (SUCCEEDED(hr))
		{
			LoadString(hResource, IDS_CHANGE_PIN, tmpStr, TMP_STR_LEN);
			hr = SHStrDupW(tmpStr, &_rgFieldStrings[PFCPI_LABEL]);
		}
		if (SUCCEEDED(hr))
		{
			LoadString(hResource, IDS_USERNAME, tmpStr, TMP_STR_LEN);
			hr = SHStrDupW(tmpStr, &_rgFieldStrings[PFCPI_USERNAME]);
		}
		if (SUCCEEDED(hr))
		{
			LoadString(hResource, IDS_OLD_OTC, tmpStr, TMP_STR_LEN);
			hr = SHStrDupW(tmpStr, &_rgFieldStrings[PFCPI_OLD_OTC]);
		}
		if (SUCCEEDED(hr))
		{
			LoadString(hResource, IDS_NEW_OTC1, tmpStr, TMP_STR_LEN);
			hr = SHStrDupW(tmpStr, &_rgFieldStrings[PFCPI_NEW_OTC1]);
		}
		if (SUCCEEDED(hr))
		{
			LoadString(hResource, IDS_NEW_OTC2, tmpStr, TMP_STR_LEN);
			hr = SHStrDupW(tmpStr, &_rgFieldStrings[PFCPI_NEW_OTC2]);
		}
		if (SUCCEEDED(hr))
		{
			LoadString(hResource, IDS_SUBMIT, tmpStr, TMP_STR_LEN);
			hr = SHStrDupW(tmpStr, &_rgFieldStrings[PFCPI_SUBMIT]);
		}
	} else {
		if (SUCCEEDED(hr))
		{
			LoadString(hResource, IDS_OTC, tmpStr, TMP_STR_LEN);
			hr = SHStrDupW(tmpStr, &_rgFieldStrings[PFI_OTC]);
		}
	}
	if (SUCCEEDED(hr))
	{
		LoadString(hResource, IDS_START_SESSION, tmpStr, TMP_STR_LEN);
		hr = SHStrDupW(tmpStr, &_rgFieldStrings[PFI_TURING_BUTTON]);
		if (SUCCEEDED(hr))
		{
			hr = SHStrDupW(tmpStr, &_pRequestImage);
			_pLinkText = _pRequestImage;
		}
	}
	if (SUCCEEDED(hr))
	{
		LoadString(hResource, IDS_REQUEST_STRING, tmpStr, TMP_STR_LEN);
		hr = SHStrDupW(tmpStr, &_rgFieldStrings[PFI_DEMAND_BUTTON]);
	}
	if (SUCCEEDED(hr))
	{
		LoadString(hResource, IDS_SESSION_PENDING, tmpStr, TMP_STR_LEN);
		hr = SHStrDupW(tmpStr, &_pCredWaiting);
	}
	if (SUCCEEDED(hr))
	{
		LoadString(hResource, IDS_PINSAFE_UNAVAILABLE_TITLE, tmpStr, TMP_STR_LEN);
		hr = SHStrDupW(tmpStr, &_pPINsafeUnavailable);
	}

    return hr;
}

// LogonUI calls this in order to give us a callback in case we need to notify it of 
// anything. We'll also provide it to the wrapped credential.
HRESULT PINsafeLoginCredential::Advise(
    ICredentialProviderCredentialEvents* pcpce
    )
{
    HRESULT hr = S_OK;

    // Grab one for ourself.
    if (_pCredProvCredentialEvents != NULL)
    {
        _pCredProvCredentialEvents->Release();
    }
    _pCredProvCredentialEvents = pcpce;
    _pCredProvCredentialEvents->AddRef();

    // Pass it along for the wrapped credential to use.

	if (_provider->GetDomainInfoIndex() > 0 && _logonDomainInfo != NULL) {
		_pCredProvCredentialEvents->SetFieldString(this, _provider->GetDomainInfoIndex(), _logonDomainInfo->String());
//		_pCredProvCredentialEvents->SetFieldState(this, _provider->GetDomainInfoIndex(), CPFS_DISPLAY_IN_SELECTED_TILE);
	} else if (_pWrappedCredential != NULL) {
        hr = _pWrappedCredential->Advise(pcpce);
	}

	// Get a handle to the credential window.
	_pCredProvCredentialEvents->OnCreatingWindow(&_hCredWindow);

	if (IsPasswordChangeRequired()) {
		//_pCredProvCredentialEvents->SetFieldString(this, _provider->GetUsernameIndex(), _provider->GetLastUsername());
		_pCredProvCredentialEvents->SetFieldString(this, _provider->GetPasswordIndex(), _provider->GetLastPassword());
		_pCredProvCredentialEvents->SetFieldState(this, _provider->GetNewPasswordIndex(), CPFS_DISPLAY_IN_SELECTED_TILE);
		_pCredProvCredentialEvents->SetFieldState(this, _provider->GetConfirmPasswordIndex(), CPFS_DISPLAY_IN_SELECTED_TILE);
		// Hide the other fields.
		_pCredProvCredentialEvents->SetFieldState(this, _provider->GetFieldIndex(PFI_OTC), CPFS_HIDDEN);
		int idx = _provider->GetFieldIndex(PFI_TURING_BUTTON);
		if (idx >= 0) {
			_pCredProvCredentialEvents->SetFieldState(this, idx, CPFS_HIDDEN);
		}
		idx = _provider->GetFieldIndex(PFI_DEMAND_BUTTON);
		if (idx >= 0) {
			_pCredProvCredentialEvents->SetFieldState(this, idx, CPFS_HIDDEN);
		}
	} else if (_provider->IsPasswordChangeComplete()) {
		_pCredProvCredentialEvents->SetFieldInteractiveState(this, _provider->GetUsernameIndex(), CPFIS_NONE);
		ExpandableString* fullUsername = new ExpandableString(_provider->GetLastDomain());
		if (!fullUsername->IsEmpty()) {
			fullUsername->Append(L"\\");
		}
		fullUsername->Append(_provider->GetLastUsername());
		_pCredProvCredentialEvents->SetFieldString(this, _provider->GetUsernameIndex(), fullUsername->String());
		delete fullUsername;
		_pCredProvCredentialEvents->SetFieldInteractiveState(this, _provider->GetPasswordIndex(), CPFIS_NONE);
		_pCredProvCredentialEvents->SetFieldString(this, _provider->GetPasswordIndex(), _provider->GetLastPassword());
		_pCredProvCredentialEvents->SetFieldState(this, _provider->GetNewPasswordIndex(), CPFS_HIDDEN);
		_pCredProvCredentialEvents->SetFieldState(this, _provider->GetConfirmPasswordIndex(), CPFS_HIDDEN);
		_pCredProvCredentialEvents->SetFieldState(this, _provider->GetFieldIndex(PFI_OTC), CPFS_DISPLAY_IN_SELECTED_TILE);
		int idx = _provider->GetFieldIndex(PFI_TURING_BUTTON);
		if (idx >= 0) {
			_pCredProvCredentialEvents->SetFieldState(this, idx, CPFS_DISPLAY_IN_SELECTED_TILE);
		}
		idx = _provider->GetFieldIndex(PFI_DEMAND_BUTTON);
		if (idx >= 0) {
			_pCredProvCredentialEvents->SetFieldState(this, idx, CPFS_DISPLAY_IN_SELECTED_TILE);
		}
	}

    return hr;
}

// LogonUI calls this to tell us to release the callback. 
// We'll also provide it to the wrapped credential.
HRESULT PINsafeLoginCredential::UnAdvise()
{
    HRESULT hr = S_OK;

    // Release our reference.
    if (_pCredProvCredentialEvents)
    {
        _pCredProvCredentialEvents->Release();
    }
    _pCredProvCredentialEvents = NULL;

    // Allow the wrapped credential to release its reference.
    if (_pWrappedCredential != NULL)
    {
        hr = _pWrappedCredential->UnAdvise();
    }

    return hr;
}

// LogonUI calls this function when our tile is selected (zoomed)
// If you simply want fields to show/hide based on the selected state,
// there's no need to do anything here - you can set that up in the 
// field definitions. In fact, we're just going to hand it off to the
// wrapped credential in case it wants to do something.
HRESULT PINsafeLoginCredential::SetSelected(BOOL* pbAutoLogon)  
{
    HRESULT hr = E_UNEXPECTED;

	if (_isChangePIN) {
		hr = S_OK;
	} else if (_pWrappedCredential != NULL) {
		// Trap autologon.
		BOOL dummy;
        hr = _pWrappedCredential->SetSelected(&dummy);
    }
	_credentialSelected = TRUE;

	ClearCredentials();
	// Allow auto-logon after password change.
	*pbAutoLogon = (_provider->IsPasswordChangeComplete());

    return hr;
}

// Clear out the username, password and OTC fields.
void PINsafeLoginCredential::ClearCredentials()
{
	if (_isChangePIN) {
		// Clear username and all OTCs
		for (DWORD idx=0; idx<_provider->GetAdditionalFieldCount(); idx++) {
			int offset = GetFieldOffset(idx);
			if (offset == PFCPI_OLD_OTC || offset == PFCPI_NEW_OTC1 || offset == PFCPI_NEW_OTC2) {
				_pCredProvCredentialEvents->SetFieldString(this, idx, L"");
			}
		}
		if (_username != NULL) _username->Clear();
		if (_currentOTC != NULL) _currentOTC->Clear();
		if (_newOTC != NULL) _newOTC->Clear();
		if (_newOTC2 != NULL) _newOTC2->Clear();
	} else {
		_pCredProvCredentialEvents->SetFieldString(this, _provider->GetPasswordIndex(), L"");
		if (_username != NULL) _username->Clear();
		if (_password != NULL) _password->Clear();
		if (_currentOTC != NULL) _currentOTC->Clear();
		// Clear OTC
		for (DWORD idx=_dwWrappedDescriptorCount; idx<_dwWrappedDescriptorCount + _provider->GetAdditionalFieldCount(); idx++) {
			if (GetFieldOffset(idx) == PFI_OTC) {
				_pCredProvCredentialEvents->SetFieldString(this, idx, L"");
				break;
			}
		}
	}
}

// Similarly to SetSelected, LogonUI calls this when your tile was selected
// and now no longer is. We'll let the wrapped credential do anything it needs.
HRESULT PINsafeLoginCredential::SetDeselected()
{
    HRESULT hr = E_UNEXPECTED;

	// Make sure you close down the TURing image.
	if (_turingThread != NULL) {
		_turingThread->Terminate();
		_turingThread = NULL;
	}

    if (!_isChangePIN && _pWrappedCredential != NULL)
    {
        hr = _pWrappedCredential->SetDeselected();
    }

	_credentialSelected = FALSE;

    return hr;
}

// Get info for a particular field of a tile. Called by logonUI to get information to 
// display the tile. We'll check to see if it's for us or the wrapped credential, and then
// handle or route it as appropriate.
HRESULT PINsafeLoginCredential::GetFieldState(
    DWORD dwFieldID,
    CREDENTIAL_PROVIDER_FIELD_STATE* pcpfs,
    CREDENTIAL_PROVIDER_FIELD_INTERACTIVE_STATE* pcpfis
    )
{
    HRESULT hr = E_UNEXPECTED;

	if (_isChangePIN) {
		FIELD_STATE_PAIR *pfsp = _LookupLocalFieldStatePair(dwFieldID);
        // If the field ID is valid, give it info it needs.
        if (pfsp != NULL) {
			*pcpfs = pfsp->cpfs;
            *pcpfis = pfsp->cpfis;
            hr = S_OK;
        } else {
			hr = E_INVALIDARG;
        }
	} else if (_pWrappedCredential != NULL) {
		// Make sure we have a wrapped credential.
        // Validate parameters.
        if ((pcpfs != NULL) && (pcpfis != NULL))
        {
            // If the field is in the wrapped credential, hand it off.
            if (_IsFieldInWrappedCredential(dwFieldID))
            {
                hr = _pWrappedCredential->GetFieldState(dwFieldID, pcpfs, pcpfis);
            }
            // Otherwise, we need to see if it's one of ours.
            else
            {
                FIELD_STATE_PAIR *pfsp = _LookupLocalFieldStatePair(dwFieldID);
                // If the field ID is valid, give it info it needs.
                if (pfsp != NULL)
                {
                    *pcpfs = pfsp->cpfs;
                    *pcpfis = pfsp->cpfis;
					// If we're changing the password, don't show our fields.
					if (IsPasswordChangeRequired()) {
						*pcpfs = CPFS_HIDDEN;
					}
					if (_isCredUI) {
						// Hide the links because they don't work.
						int fieldIdx = GetFieldOffset(dwFieldID);
						if (fieldIdx == PFI_TURING_BUTTON || fieldIdx == PFI_DEMAND_BUTTON) {
							*pcpfs = CPFS_HIDDEN;
						}
					}

                    hr = S_OK;
                }
                else
                {
                    hr = E_INVALIDARG;
                }
            }
        }
        else
        {
            hr = E_INVALIDARG;
        }
    }
    return hr;
}

// Sets ppwsz to the string value of the field at the index dwFieldID. We'll check to see if 
// it's for us or the wrapped credential, and then handle or route it as appropriate.
HRESULT PINsafeLoginCredential::GetStringValue(
    DWORD dwFieldID, 
    PWSTR* ppwsz
    )
{
    HRESULT hr = E_UNEXPECTED;

    // Make sure we have a wrapped credential.
    if (_pWrappedCredential != NULL)
    {
        // If this field belongs to the wrapped credential, hand it off.
		int offset = GetFieldOffset(dwFieldID);
        if (offset < 0) {
			hr = _pWrappedCredential->GetStringValue(dwFieldID, ppwsz);
        }
        // Otherwise determine if we need to handle it.
		else
		{
			switch (offset) {
			case PFI_OTC:
			case PFCPI_OLD_OTC:
				// Get the OTC.
				if (_currentOTC == NULL) {
					SHStrDup(L"", ppwsz);
				} else {
					_currentOTC->CopyToString(ppwsz);
				}
				hr = S_OK;
				break;

			case PFCPI_USERNAME:
				// Get the username
				if (_username == NULL) {
					SHStrDup(L"", ppwsz);
				} else {
					_username->CopyToString(ppwsz);
				}
				hr = S_OK;
				break;

			case PFCPI_NEW_OTC1:
				// Get the first OTC
				if (_newOTC == NULL) {
					SHStrDup(L"", ppwsz);
				} else {
					_newOTC->CopyToString(ppwsz);
				}
				hr = S_OK;
				break;

			case PFCPI_NEW_OTC2:
				// Get the second OTC
				if (_newOTC2 == NULL) {
					SHStrDup(L"", ppwsz);
				} else {
					_newOTC2->CopyToString(ppwsz);
				}
				hr = S_OK;
				break;

			case PFI_TURING_BUTTON:
				hr = SHStrDupW(_pLinkText, ppwsz);
				break;

			default:
				FIELD_STATE_PAIR *pfsp = _LookupLocalFieldStatePair(dwFieldID);
				if (pfsp != NULL)
				{
					hr = SHStrDupW(_rgFieldStrings[offset], ppwsz);
				}
				else
				{
					hr = E_INVALIDARG;
				}
			}
		}
    }
    return hr;
}

// Returns the number of items to be included in the combobox (pcItems), as well as the 
// currently selected item (pdwSelectedItem). We'll check to see if it's for us or the 
// wrapped credential, and then handle or route it as appropriate.
HRESULT PINsafeLoginCredential::GetComboBoxValueCount(
    DWORD dwFieldID, 
    DWORD* pcItems, 
    DWORD* pdwSelectedItem
    )
{
    HRESULT hr = E_UNEXPECTED;

    // Make sure we have a wrapped credential.
    if (_pWrappedCredential != NULL)
    {
        // If this field belongs to the wrapped credential, hand it off.
        if (_IsFieldInWrappedCredential(dwFieldID))
        {
            hr = _pWrappedCredential->GetComboBoxValueCount(dwFieldID, pcItems, pdwSelectedItem);
        }
        // Otherwise determine if we need to handle it.
        else
        {
            if (FALSE) // TODO
            {
                hr = S_OK;
            }
            else
            {
                hr = E_INVALIDARG;
            }
        }
    }

    return hr;
}

// Called iteratively to fill the combobox with the string (ppwszItem) at index dwItem.
// We'll check to see if it's for us or the wrapped credential, and then handle or route 
// it as appropriate.
HRESULT PINsafeLoginCredential::GetComboBoxValueAt(
    DWORD dwFieldID, 
    DWORD dwItem,
    PWSTR* ppwszItem
    )
{
    HRESULT hr = E_UNEXPECTED;

    // Make sure we have a wrapped credential.
    if (_pWrappedCredential != NULL)
    {
        // If this field belongs to the wrapped credential, hand it off.
        if (_IsFieldInWrappedCredential(dwFieldID))
        {
            hr = _pWrappedCredential->GetComboBoxValueAt(dwFieldID, dwItem, ppwszItem);
        }
        // Otherwise determine if we need to handle it.
        else
        {
            hr = E_INVALIDARG;
        }
    }

    return hr;
}

// Called when the user changes the selected item in the combobox. We'll check to see if 
// it's for us or the wrapped credential, and then handle or route it as appropriate.
HRESULT PINsafeLoginCredential::SetComboBoxSelectedValue(
    DWORD dwFieldID,
    DWORD dwSelectedItem
    )
{
    HRESULT hr = E_UNEXPECTED;

    // Make sure we have a wrapped credential.
    if (_pWrappedCredential != NULL)
    {
        // If this field belongs to the wrapped credential, hand it off.
        if (_IsFieldInWrappedCredential(dwFieldID))
        {
            hr = _pWrappedCredential->SetComboBoxSelectedValue(dwFieldID, dwSelectedItem);
        }
        // Otherwise determine if we need to handle it.
        else
        {
            hr = E_INVALIDARG;
        }
    }

    return hr;
}

//------------- 
// The following methods are for logonUI to get the values of various UI elements and 
// then communicate to the credential about what the user did in that field. Even though
// we don't offer these field types ourselves, we need to pass along the request to the
// wrapped credential.

HRESULT PINsafeLoginCredential::GetBitmapValue(
    DWORD dwFieldID, 
    HBITMAP* phbmp
    )
{
    HRESULT hr = E_UNEXPECTED;

	if (_pWrappedCredential != NULL) {
		hr = _pWrappedCredential->GetBitmapValue(dwFieldID, phbmp);
    }

    return hr;
}

HRESULT PINsafeLoginCredential::GetSubmitButtonValue(
    DWORD dwFieldID,
    DWORD* pdwAdjacentTo
    )
{
    HRESULT hr = E_UNEXPECTED;

	if (_isChangePIN) {
		// Put the button next to the Confirm New OTC field.
		*pdwAdjacentTo = PFCPI_NEW_OTC2;
		hr = 0;
	} else if (_pWrappedCredential != NULL)
    {
        hr = _pWrappedCredential->GetSubmitButtonValue(dwFieldID, pdwAdjacentTo);
    }

    return hr;
}

HRESULT PINsafeLoginCredential::SetStringValue(
    DWORD dwFieldID,
    PCWSTR pwz
    )
{
    HRESULT hr = E_UNEXPECTED;
	DWORD offset = GetFieldOffset(dwFieldID);

	switch (offset) {
	case PFI_OTC:
	case PFCPI_OLD_OTC:
		// Get the one-time code.
		if (_currentOTC == NULL) {
			_currentOTC = new ExpandableString(pwz);
		} else {
			_currentOTC->Set(pwz);
		}
		break;

	case PFCPI_USERNAME:
		// Get the username (change PIN only).
		if (_username == NULL) {
			_username = new ExpandableString(pwz);
		} else {
			_username->Set(pwz);
		}
		break;

	case PFCPI_NEW_OTC1:
		// Get the new one-time code.
		if (_newOTC == NULL) {
			_newOTC = new ExpandableString(pwz);
		} else {
			_newOTC->Set(pwz);
		}
		break;

	case PFCPI_NEW_OTC2:
		// Get the new one-time code.
		if (_newOTC2 == NULL) {
			_newOTC2 = new ExpandableString(pwz);
		} else {
			_newOTC2->Set(pwz);
		}
		break;

	case PFCPI_SUBMIT:
		// Setting string for Submit button
		break;

	default:
		if (!_isChangePIN && _pWrappedCredential != NULL) {
	        hr = _pWrappedCredential->SetStringValue(dwFieldID, pwz);
			if (dwFieldID == _provider->GetUsernameIndex()) {
				LPCWSTR logonPrompt = _provider->GetResourceString(IDS_LOG_ON_TO);
				if (_logonDomainInfo == NULL) {
					_logonDomainInfo = new ExpandableString(logonPrompt);
				} else {
					_logonDomainInfo->Set(logonPrompt);
				}
				// Check if this is a local or domain user.
				LPCWSTR slashPos = wcschr(pwz, L'\\');
				if (slashPos != NULL) {
					// Force the domain name.
					LPWSTR domain;
					SHStrDup(pwz, &domain);
					slashPos = wcschr(domain, L'\\');
					_logonDomainInfo->Append(domain, (size_t)(slashPos-domain));
					CoTaskMemFree(domain);
				} else if (wcsicmp(pwz, L"administrator") == 0) {
					// AFAICT, only "administrator" defaults to local account.
					_logonDomainInfo->Append(_computerName->String());
				} else {
					_logonDomainInfo->Append(_domain->String());
				}
				if (_provider->GetDomainInfoIndex() > 0) {
					SetStringValue(_provider->GetDomainInfoIndex(), _logonDomainInfo->String());
				}
				ExpandableString* oldUsername = _username;
				if (_username == NULL) {
					_username = new ExpandableString(pwz);
				} else {
					_username->Set(pwz);
				}
			} else if (dwFieldID == _provider->GetPasswordIndex()) {
				if (_password == NULL) {
					_password = new ExpandableString(pwz);
				} else {
					_password->Set(pwz);
				}
			} else if (dwFieldID == _provider->GetNewPasswordIndex()) {
				if (_newPassword == NULL) {
					_newPassword = new ExpandableString(pwz);
				} else {
					_newPassword->Set(pwz);
				}
			} else if (dwFieldID == _provider->GetConfirmPasswordIndex()) {
				if (_newPassword2 == NULL) {
					_newPassword2 = new ExpandableString(pwz);
				} else {
					_newPassword2->Set(pwz);
				}
			}
		}
	}

    return hr;
}

HRESULT PINsafeLoginCredential::GetCheckboxValue(
    DWORD dwFieldID, 
    BOOL* pbChecked,
    PWSTR* ppwszLabel
    )
{
    HRESULT hr = E_UNEXPECTED;

    if (_pWrappedCredential != NULL)
    {
        if (_IsFieldInWrappedCredential(dwFieldID))
        {
            hr = _pWrappedCredential->GetCheckboxValue(dwFieldID, pbChecked, ppwszLabel);
        }
    }

    return hr;
}

HRESULT PINsafeLoginCredential::SetCheckboxValue(
    DWORD dwFieldID, 
    BOOL bChecked
    )
{
    HRESULT hr = E_UNEXPECTED;

    if (_pWrappedCredential != NULL)
    {
        hr = _pWrappedCredential->SetCheckboxValue(dwFieldID, bChecked);
    }

    return hr;
}

void PINsafeLoginCredential::RequestSecurityString() {
	CheckUsername();
	if (_username->IsEmpty()) {
		return;
	}
	PINsafeConnect* client = new PINsafeConnect();
	client->RequestSecurityString(_username->String());
}

void PINsafeLoginCredential::ShowPINsafeImage() {

	ShowBusyCursor();
	_isShowingImage = TRUE;
	if (!CheckUsername()) {
		LPCWSTR msg = _provider->GetResourceString(IDS_NO_USERNAME);
		LPCWSTR title = _provider->GetResourceString(IDS_NO_USERNAME_TITLE);
		MessageBox(_hCredWindow, msg, title, MB_ICONERROR | MB_OK);
		CoTaskMemFree((LPVOID)msg);
		CoTaskMemFree((LPVOID)title);
		_isShowingImage = FALSE;
		return;
	}

	// Change the prompt to indicate that we are looking for PINsafe.
	_pLinkText = _pCredWaiting;
	_pCredProvCredentialEvents->SetFieldString(this, _turingBtnIndex, _pCredWaiting);
	UpdateWindow(_hCredWindow);

	// Start the thread to show the TURing image.
	if (_turingThread == NULL || _dialogTerminated) {
		_dialogTerminated = FALSE;
		// Display the dialog
		_turingThread = (TuringThread*)AfxBeginThread(RUNTIME_CLASS(TuringThread));
		_turingThread->SetCredential(this);
	} else {
		if (!_turingThread->Visible()) {
			_turingThread->SetCredential(this);
			_turingThread->InitInstance();
		}
	}

	_useLocal = _settings->UseLocalAuth();
	PINsafeConnect* client = new PINsafeConnect();
	// Don't check PINsafe connectivity unless we need it
	if ((!_useLocal && _settings->AllowLocalAuth()) || _settings->AllowDefaultLogin()) {
		_isPINsafeAvailable = client->Ping();
	}
	if (!_isPINsafeAvailable) {
		// PINsafe not available.
		_pLinkText = _pPINsafeUnavailable;
		_pCredProvCredentialEvents->SetFieldString(this, _turingBtnIndex, _pPINsafeUnavailable);
		ShowNormalCursor();
		UpdateWindow(_hCredWindow);
		// If we're not using local authentication, close the TURing window.
		if (!_settings->AllowLocalAuth()) {
			_turingThread->Terminate();
			_turingThread = NULL;
		}
	}
	if (!_useLocal && _settings->AllowLocalAuth() && !_isPINsafeAvailable) {
		// Use local if PINsafe server not available
		_useLocal = TRUE;
	}

	if (_useLocal) {
		_turingLocal = new CTuringGenerator();
		_turingThread->SetLocalImage(_turingLocal, _username->String(), g_hinst);
	} else if (!_isPINsafeAvailable) {
		LPCWSTR msg = _provider->GetResourceString(IDS_PINSAFE_UNAVAILABLE);
		LPCWSTR title = _provider->GetResourceString(IDS_PINSAFE_UNAVAILABLE_TITLE);
		MessageBox(_hCredWindow, msg, title, MB_OK | MB_ICONERROR);
		CoTaskMemFree((LPVOID)msg);
		CoTaskMemFree((LPVOID)title);
	} else {
		IPicture* _turingImage = GetTuringImage(client);
		if (_turingImage == NULL) {
			LogError(L"Failed to get TURing image");
		} else {
			_turingThread->SetImage(_turingImage);
		}
	}

	delete client;
}

void PINsafeLoginCredential::CheckUsernameChanged() {
	if (_settings->IgnoreDomain()) {
		//LogInfo(L"Checking username %s", _username->String());
		// Strip off domain info
		int ptr = _username->Locate(L'\\');
		if (ptr >= 0) {
			_username->TruncateLeft(ptr + 1);
		} else {
			ptr = _username->Locate(L'@');
			if (ptr >= 0) {
				_username->Truncate(ptr);
			}
		}
	}
	if ((_oldUsername == NULL && _username->Length() > 0) || (_oldUsername != NULL) && _username->Compare(*_oldUsername)) {
		_useLocal = CheckUseLocal();
		if (_allowUnknownUsers && _username != NULL) {
			_userExists = UserExists(_useLocal);
		}
	}
	_oldUsername = new ExpandableString(*_username);
}

BOOL PINsafeLoginCredential::UserExists(BOOL useLocal) {
	BOOL userExists;
	LPWSTR name = new WCHAR[_username->Length() + 1];
	_username->CopyToString(&name);
	if (useLocal) {
		// If local authentication is active, check if user is known locally.
		userExists = CredentialCache::LocalUserExists(name);
		//LogInfo(L"Local user %s exists? %d", name, userExists);
	} else {
		// If PINsafe is active, check if the user is known to PINsafe.
		PINsafeConnect* connect = new PINsafeConnect();
		userExists = connect->UserExists(name);
		//LogInfo(L"User %s exists? %d", name, userExists);
	}
	return userExists;
}

BOOL PINsafeLoginCredential::CheckUsername() {
	if (!_isChangePIN) {
		int uidx = _provider->GetUsernameIndex();
		if (uidx >= 0) {
			if (_username == NULL || _username->IsEmpty()) {
				// Check if the username is pre-set.
				LPWSTR temp;
				if (SUCCEEDED(GetStringValue(uidx, &temp))) {
					if (_username == NULL) {
						_username = new ExpandableString(temp);
					} else {
						_username->Set(temp);
					}
				}
			}
			if (_username == NULL || _username->IsEmpty()) {
				LogError(L"No username found");
				return FALSE;
			}
			CheckUsernameChanged();
		} else {
			LogError(L"Unable to identify username field");
			return FALSE;
		}
	}
	return (_username != NULL && !_username->IsEmpty());
}

IPicture* PINsafeLoginCredential::GetTuringImage(PINsafeConnect* client) {
	IPicture* image;

	if (_username == NULL || _username->IsEmpty()) {
		LogError(L"No username - using blank image");
		return NULL;
	}

	image = client->GetSCImageByUsername(_username->String());

	if (image == NULL) {
		LogError(L"Failed to get TURing image");
		_isPINsafeAvailable = FALSE;
		return NULL;
	}

	return image;
}

HRESULT PINsafeLoginCredential::CommandLinkClicked(DWORD dwFieldID)
{
	if (!_isChangePIN && _pWrappedCredential != NULL && dwFieldID < _dwWrappedDescriptorCount) {
		return _pWrappedCredential->CommandLinkClicked(dwFieldID);
	}
    HRESULT hr = E_UNEXPECTED;

	switch (GetFieldOffset(dwFieldID)) {
		case PFI_TURING_BUTTON:
			if (!_isShowingImage) {
				_turingBtnIndex = dwFieldID;
				// We clicked the TURing button
				ShowPINsafeImage();
				hr = S_OK;
			}
			break;
		case PFI_DEMAND_BUTTON:
			RequestSecurityString();
			hr = S_OK;
			break;
	}

    return hr;
}
//------ end of methods for controls we don't have ourselves ----//


//
// Collect the username and password into a serialized credential for the correct usage scenario 
// LogonUI then passes these credentials back to the system to log on.
//
HRESULT PINsafeLoginCredential::GetSerialization(
    CREDENTIAL_PROVIDER_GET_SERIALIZATION_RESPONSE* pcpgsr,
    CREDENTIAL_PROVIDER_CREDENTIAL_SERIALIZATION* pcpcs, 
    PWSTR* ppwszOptionalStatusText, 
    CREDENTIAL_PROVIDER_STATUS_ICON* pcpsiOptionalStatusIcon
    )
{
    HRESULT hr = E_UNEXPECTED;
	LPCWSTR msg;
	LPCWSTR title;

	if (_turingThread != NULL) {
		_turingThread->Terminate();
		_turingThread = NULL;
	}

	PINsafeConnect *client = new PINsafeConnect();

	if (_isChangePIN) {
		if (_username == NULL || _username->IsEmpty() || _currentOTC == NULL || _currentOTC->IsEmpty() || _newOTC == NULL || _newOTC->IsEmpty() || _newOTC2 == NULL || _newOTC2->IsEmpty()) {
			msg = _provider->GetResourceString(IDS_MISSING_VALUES);
			title = _provider->GetResourceString(IDS_MISSING_VALUES_TITLE);
			MessageBox(_hCredWindow, msg, title, MB_OK | MB_ICONERROR);
			CoTaskMemFree((LPVOID)msg);
			CoTaskMemFree((LPVOID)title);
			*pcpgsr = CPGSR_NO_CREDENTIAL_NOT_FINISHED;
			return E_FAIL;
		}
		if (_newOTC->Compare(*_newOTC2) != 0) {
			msg = _provider->GetResourceString(IDS_OTC_MISMATCH);
			title = _provider->GetResourceString(IDS_OTC_MISMATCH_TITLE);
			MessageBox(_hCredWindow, msg, title, MB_OK | MB_ICONERROR);
			CoTaskMemFree((LPVOID)msg);
			CoTaskMemFree((LPVOID)title);
			return E_FAIL;
		}
		if (client->ChangePIN(_username->String(), L"", L"", _currentOTC->String(), _newOTC->String())) {
			msg = _provider->GetResourceString(IDS_CHANGE_PIN_SUCCESS);
			title = _provider->GetResourceString(IDS_CHANGE_PIN_TITLE);
			MessageBox(_hCredWindow, msg, title, MB_OK | MB_ICONINFORMATION);
			CoTaskMemFree((LPVOID)msg);
			CoTaskMemFree((LPVOID)title);
			*pcpgsr = CPGSR_RETURN_CREDENTIAL_FINISHED;
			hr = S_OK;
		} else {
			msg = _provider->GetResourceString(IDS_CHANGE_PIN_FAILED);
			title = _provider->GetResourceString(IDS_CHANGE_PIN_TITLE);
			MessageBox(_hCredWindow, msg, title, MB_OK | MB_ICONERROR);
			CoTaskMemFree((LPVOID)msg);
			CoTaskMemFree((LPVOID)title);
			*pcpgsr = CPGSR_NO_CREDENTIAL_NOT_FINISHED;
			hr = E_FAIL;
		}
		if (hr == E_FAIL) {
			// We might be trying again, so clear out the invalid credentials.
			for (DWORD idx=0; idx<_provider->GetAdditionalFieldCount(); idx++) {
				int offset = GetFieldOffset(idx);
				if (offset == PFCPI_OLD_OTC || offset == PFCPI_NEW_OTC1 || offset == PFCPI_NEW_OTC2) {
					SetStringValue(idx, L"");
				}
			}
		}
		return hr;
	} else if (IsPasswordChangeRequired()) {
		// Serialize a change password request.
		KERB_CHANGEPASSWORD_REQUEST cpr;
		cpr.MessageType = KerbChangePasswordMessage;
		LPWSTR domain, username, oldPassword, newPassword, newPassword2;
		SHStrDup(_provider->GetLastDomain(), &domain);
		UnicodeStringInitWithString(domain, &cpr.DomainName);
		SHStrDup(_provider->GetLastUsername(), &username);
		UnicodeStringInitWithString(username, &cpr.AccountName);
		SHStrDup(_provider->GetLastPassword(), &oldPassword);
		UnicodeStringInitWithString(oldPassword, &cpr.OldPassword);
		SHStrDup(_newPassword->String(), &newPassword);
		SHStrDup(_newPassword2->String(), &newPassword2);
		if (wcscmp(newPassword, newPassword2) != 0) {
			// Passwords don't match - try again.
			CoTaskMemFree(newPassword);
			CoTaskMemFree(newPassword2);
			msg = _provider->GetResourceString(IDS_PASSWORD_MISMATCH);
			title = _provider->GetResourceString(IDS_PASSWORD_MISMATCH_TITLE);
			MessageBox(_hCredWindow, msg, title, MB_OK | MB_ICONERROR);
			*pcpgsr = CPGSR_NO_CREDENTIAL_NOT_FINISHED;
			return E_FAIL;
		}
		CoTaskMemFree(newPassword2);
		UnicodeStringInitWithString(newPassword, &cpr.NewPassword);
		cpr.Impersonating = FALSE;
		hr = KerbChangePasswordPack(cpr, &pcpcs->rgbSerialization, &pcpcs->cbSerialization);
		if (SUCCEEDED(hr)) {
			ULONG ulAuthPackage;
			hr = RetrieveNegotiateAuthPackage(&ulAuthPackage);
			if (SUCCEEDED(hr))
			{
				pcpcs->ulAuthenticationPackage = ulAuthPackage;
				pcpcs->clsidCredentialProvider = CLSID_PasswordCredentialProvider;
		 
				*pcpgsr = CPGSR_RETURN_CREDENTIAL_FINISHED;
			}
		}

		// OK, the structure has been serialized - we can get rid of it now.
		CoTaskMemFree(domain);
		CoTaskMemFree(username);
		CoTaskMemFree(oldPassword);
		// Save the new password.
		_provider->SaveLastCredentials(NULL, newPassword);
		CoTaskMemFree(newPassword);
		return hr;
	} else if (_provider->IsPasswordChangeComplete()) {
		// If we've just changed password, re-use the new password to authenticate.
		KERB_INTERACTIVE_LOGON kil;
		kil.MessageType = KerbInteractiveLogon;
		LPWSTR domain, username, password;
		SHStrDup(_provider->GetLastDomain(), &domain);
		SHStrDup(_provider->GetLastUsername(), &username);
		SHStrDup(_provider->GetLastPassword(), &password);
		UnicodeStringInitWithString(domain, &kil.LogonDomainName);
		UnicodeStringInitWithString(username, &kil.UserName);
		UnicodeStringInitWithString(password, &kil.Password);
		hr = KerbInteractiveLogonPack(kil, &pcpcs->rgbSerialization, &pcpcs->cbSerialization);
		if (SUCCEEDED(hr)) {
			ULONG ulAuthPackage;
			hr = RetrieveNegotiateAuthPackage(&ulAuthPackage);
			if (SUCCEEDED(hr))
			{
				pcpcs->ulAuthenticationPackage = ulAuthPackage;
				pcpcs->clsidCredentialProvider = CLSID_PasswordCredentialProvider;
		 
				*pcpgsr = CPGSR_RETURN_CREDENTIAL_FINISHED;
			}
		}
		CoTaskMemFree(domain);
		CoTaskMemFree(username);
		CoTaskMemFree(password);
		return hr;
	} else if (_pWrappedCredential != NULL) {
		LoginState result = LOGIN_UNKNOWN;
		_useLocal = CheckUseLocal();
		BOOL isPINsafeRequired = TRUE;
		BOOL checkUser = CheckUsername();
		if (_allowUnknownUsers) {
			_userExists = UserExists(_useLocal);
			if (!_userExists) {
				isPINsafeRequired = FALSE;
			}
		}
		if (_settings->LoginDisabled() || (_allowUnknownUsers && !_userExists)) {
			//LogInfo(L"PINsafe authentication by-passed");
			result = LOGIN_PASSED; // No PINsafe required.
		}
		else if (!_settings->LoginLocal())
		{
			BOOL remote = GetSystemMetrics(SM_REMOTESESSION);
			if (!remote)
			{
				result = LOGIN_PASSED; // No PINsafe for local login.
			}
		}
		if (result != LOGIN_PASSED && !_useLocal && _settings->AllowDefaultLogin()) {
			if (_isPINsafeAvailable) {
				// Double-check, in case it's stopped now.
				_isPINsafeAvailable = client->Ping();
			}
			if (!_isPINsafeAvailable) {
				// PINsafe not active
				result = LOGIN_PASSED;
			}
		}
		if (result != LOGIN_PASSED) {
			// Make sure username has been extracted.
			if (!checkUser) {
				msg = _provider->GetResourceString(IDS_NO_USERNAME);
				title = _provider->GetResourceString(IDS_NO_USERNAME_TITLE);
				MessageBox(_hCredWindow, msg, title, MB_OK | MB_ICONERROR);
				CoTaskMemFree((LPVOID)msg);
				CoTaskMemFree((LPVOID)title);
				*pcpgsr = CPGSR_NO_CREDENTIAL_NOT_FINISHED;
				hr = E_FAIL;
				result = LOGIN_INCOMPLETE;
			}
			if (_currentOTC == NULL || _currentOTC->IsEmpty()) {
				if (isPINsafeRequired) {
					msg = _provider->GetResourceString(IDS_NO_OTC);
					title = _provider->GetResourceString(IDS_NO_OTC_TITLE);
					MessageBox(_hCredWindow, msg, title, MB_OK | MB_ICONERROR);
					CoTaskMemFree((LPVOID)msg);
					CoTaskMemFree((LPVOID)title);
					*pcpgsr = CPGSR_NO_CREDENTIAL_NOT_FINISHED;
					hr = E_FAIL;
				}
				// If we're using standard authentication, OTC is not relevant.
				if (_isPINsafeAvailable || _useLocal) {
					result = LOGIN_INCOMPLETE;
				}
			}
			if (result == LOGIN_UNKNOWN) {
				if (_useLocal) {
					result = (_turingLocal == NULL) ? LOGIN_FAILED : (_turingLocal->CheckCredentials(_username->String(), L"", _currentOTC->String())) ? LOGIN_PASSED : LOGIN_FAILED;
					if (result == LOGIN_FAILED) {
						LogError(L"Local authentication failed for user %s", _username->String());
					} else {
						LogInfo(L"Local authentication succeeded for user %s", _username->String());
					}
				} else if (_isPINsafeAvailable) {
					BOOL requestStrings = _settings->AllowLocalAuth();
					result = client->Authenticate(_username->String(), L"", _currentOTC->String(), requestStrings);
					if (result == LOGIN_FAILED) {
						LogError(L"PINsafe authentication failed for user %s", _username->String());
					} else {
						LogInfo(L"PINsafe authentication succeeded for user %s", _username->String());
						if (result == LOGIN_PINCHANGE_REQUIRED) {
							msg = _provider->GetResourceString(IDS_PIN_CHANGE_REQUIRED);
							title = _provider->GetResourceString(IDS_PIN_CHANGE_REQUIRED_TITLE);
							MessageBox(_hCredWindow, msg, title, MB_OK | MB_ICONWARNING);
							CoTaskMemFree((LPVOID)msg);
							CoTaskMemFree((LPVOID)title);
							_PINChangeRequired = TRUE;
							_provider->SetPINChangeRequired(TRUE);
							// TODO: rather than just warning that PIN change is required, we should actually redisplay the change PIN page.
						} else if (result == LOGIN_PIN_EXPIRES_SOON) {
							msg = _provider->GetResourceString(IDS_PIN_CHANGE_WARNING);
							title = _provider->GetResourceString(IDS_PIN_CHANGE_WARNING_TITLE);
							MessageBox(_hCredWindow, msg, title, MB_OK | MB_ICONINFORMATION);
							CoTaskMemFree((LPVOID)msg);
							CoTaskMemFree((LPVOID)title);
						}
					}
				}
			}
		}
		if (result != LOGIN_FAILED && result != LOGIN_INCOMPLETE) {
			hr = _pWrappedCredential->GetSerialization(pcpgsr, pcpcs, ppwszOptionalStatusText, pcpsiOptionalStatusIcon);
			// If we've failed authentication, clear out the OTC and password.
			if (hr != S_OK) {
				LogError(L"Windows authentication failed for user %s. Return code is %x.", _username, hr);
				if (_pCredProvCredentialEvents != NULL) {
					if (_password != NULL) _password->Clear();
					if (_currentOTC != NULL) _currentOTC->Clear();
				}
			} else {
				// Save the serialized credentials from Windows so that we can re-use them.
				_provider->SaveLastCredentials(pcpcs, _password->String());
				result = LOGIN_PASSED;
			}
		} else if (result == LOGIN_FAILED) {
			// TODO: This isn't the best way to do it, but what I propose is to send a random string as the password.
			KERB_INTERACTIVE_LOGON kil;
			ZeroMemory(&kil, sizeof(kil));

			WCHAR wsz[MAX_COMPUTERNAME_LENGTH+1];
			DWORD cch = ARRAYSIZE(wsz);
			if (GetComputerNameW(wsz, &cch))
			{
				//
				// Initialize the UNICODE_STRINGS to share our username and password strings
				//
				hr = UnicodeStringInitWithString(wsz, &kil.LogonDomainName);

				if (SUCCEEDED(hr))
				{
					_username->ToUnicodeString(kil.UserName);

					wchar_t randPass[11];
					for (int idx=0; idx<10; idx++) {
						randPass[idx] = (rand() % 0x5D) + 0x21;
					}
					randPass[10] = 0;
					hr = UnicodeStringInitWithString(randPass, &kil.Password);

					if (SUCCEEDED(hr))
					{
						//
						// Allocate copies of, and package, the strings in a binary blob.
						//
						kil.MessageType = KerbInteractiveLogon;
						hr = KerbInteractiveLogonPack(kil, &pcpcs->rgbSerialization, &pcpcs->cbSerialization);

						if (SUCCEEDED(hr))
						{
							ULONG ulAuthPackage;
							hr = RetrieveNegotiateAuthPackage(&ulAuthPackage);
							if (SUCCEEDED(hr))
							{
								pcpcs->ulAuthenticationPackage = ulAuthPackage;
								pcpcs->clsidCredentialProvider = CLSID_PasswordCredentialProvider;
		 
								// At this point the credential has created the serialized credential used for logon
								// By setting this to CPGSR_RETURN_CREDENTIAL_FINISHED we are letting logonUI know
								// that we have all the information we need and it should attempt to submit the 
								// serialized credential.
								*pcpgsr = CPGSR_RETURN_CREDENTIAL_FINISHED;
							}
						}
					}
					CoTaskMemFree(kil.UserName.Buffer);
				}
			}
			else
			{
				DWORD dwErr = GetLastError();
				hr = HRESULT_FROM_WIN32(dwErr);
			}
			// We know we've failed authentication, so clear out the OTC and password.
			_password->Clear();
			for (DWORD idx=_dwWrappedDescriptorCount; idx<_dwWrappedDescriptorCount + _provider->GetAdditionalFieldCount(); idx++) {
				if (GetFieldOffset(idx) == PFI_OTC) {
					_currentOTC->Clear();
					break;
				}
			}
		}
    }
	delete client;
    return hr;
}

// ReportResult is completely optional. However, we will hand it off to the wrapped
// credential in case they want to handle it.
HRESULT PINsafeLoginCredential::ReportResult(
    NTSTATUS ntsStatus, 
    NTSTATUS ntsSubstatus,
    PWSTR* ppwszOptionalStatusText, 
    CREDENTIAL_PROVIDER_STATUS_ICON* pcpsiOptionalStatusIcon
    )
{
    HRESULT hr = E_UNEXPECTED;

	if (ntsStatus == STATUS_PASSWORD_MUST_CHANGE) {
		// Password change required.
		_provider->SetPasswordChangeRequired(TRUE);
	} else if (_provider->IsPasswordChangeRequired()) {
		if (ntsStatus == STATUS_SUCCESS) {
			_provider->SetPasswordChangeComplete(TRUE);
		}
	}
	if (_pWrappedCredential != NULL) {
        hr = _pWrappedCredential->ReportResult(ntsStatus, ntsSubstatus, ppwszOptionalStatusText, pcpsiOptionalStatusIcon);
    }

    return hr;
}

BOOL PINsafeLoginCredential::_IsFieldInWrappedCredential(DWORD dwFieldID)
{
	return (dwFieldID < _dwWrappedDescriptorCount);
}

FIELD_STATE_PAIR *PINsafeLoginCredential::_LookupLocalFieldStatePair(DWORD dwFieldID)
{
    // Offset into the ID to account for the wrapped fields.
    int offset = GetFieldOffset(dwFieldID);

    // If the index if valid, give it the info it wants.
    if (offset >= 0)
    {
        return &(_rgFieldStatePairs[offset]);
    }
    
    return NULL;
}

void PINsafeLoginCredential::ShowTURingImageVisible()
{
	_isShowingImage = FALSE;
	_pLinkText = _pRequestImage;
	_pCredProvCredentialEvents->SetFieldString(this, _turingBtnIndex, _pRequestImage);
	UpdateWindow(_hCredWindow);
	_isPINsafeAvailable = TRUE;
	ShowNormalCursor();
}

// VERY naughty to do it this way, but I'm not going to figure out how to include the right definitions.
#ifndef GCL_HCURSOR
#define GCL_HCURSOR         (-12)
#endif

void PINsafeLoginCredential::ShowBusyCursor()
{
	HCURSOR hBusyCursor = LoadCursor(NULL, IDC_WAIT);
	SetCursor(hBusyCursor);
	SetWindowLong(_hCredWindow, GCL_HCURSOR, (LONG)hBusyCursor);
}

void PINsafeLoginCredential::ShowNormalCursor()
{
	HCURSOR hCursor = LoadCursor(NULL, IDC_ARROW);
	SetCursor(hCursor);
	SetWindowLong(_hCredWindow, GCL_HCURSOR, (LONG)hCursor);
}

BOOL PINsafeLoginCredential::IsPasswordChangeRequired()
{ 
	return _provider->IsPasswordChangeRequired(); 
}

BOOL PINsafeLoginCredential::CheckUseLocal()
{
	if (_settings->UseLocalAuth()) {
		return TRUE;	// Always use local authentication.
	}
	// Don't check PINsafe connectivity unless we need it
	if ((!_useLocal && _settings->AllowLocalAuth()) || _settings->AllowDefaultLogin()) {
		PINsafeConnect* client = new PINsafeConnect();
		_isPINsafeAvailable = client->Ping();
		delete client;
		_useLocal = !_isPINsafeAvailable;
	}
	return _useLocal;
}