#include "PINsafeLoginFilter.h"
#include "Logging.h"
#include "Guid.h"

PINsafeLoginFilter::PINsafeLoginFilter(void)
{
	_settings = RegSettings::GetSettings();
}

PINsafeLoginFilter::~PINsafeLoginFilter(void)
{
}

HRESULT PINsafeLoginFilter::Filter(CREDENTIAL_PROVIDER_USAGE_SCENARIO cpus, DWORD dwFlags, GUID* rgclsidProviders, 
		BOOL* rgbAllow, DWORD cProviders)
{
	BOOL isRemote = GetSystemMetrics(SM_REMOTESESSION);
	BOOL disablePINsafe = _settings->LoginDisabled() || (_settings->LoginRemoteOnly() && !isRemote);
	BOOL disableDefault = !(_settings->DisableFilter() || disablePINsafe); // Don't disable default if PINsafe is also disabled.

	switch (cpus)
	{
	case CPUS_LOGON:
	case CPUS_UNLOCK_WORKSTATION:
	case CPUS_CREDUI:
		// Filters out the default Windows provider (only for Logon and Unlock scenarios)
		for (int i = 0; i < cProviders; i++)
		{
			if (disableDefault && IsEqualGUID(rgclsidProviders[i], CLSID_PasswordCredentialProvider)) {
				LogInfo(L"PINsafeLoginFilter: standard credential provider disabled");
				rgbAllow[i] = FALSE;
			}
			if (disablePINsafe && IsEqualGUID(rgclsidProviders[i], CLSID_PINsafeLoginProvider)) {
				LogInfo(L"PINsafeLoginFilter: PINsafe credential provider disabled");
				rgbAllow[i] = FALSE;
			}
		}
		return S_OK;

	case CPUS_CHANGE_PASSWORD:
		return E_NOTIMPL;

	default:
		return E_INVALIDARG;
	}
}

HRESULT PINsafeLoginFilter::UpdateRemoteCredential(const CREDENTIAL_PROVIDER_CREDENTIAL_SERIALIZATION* pcpcsIn, 
		CREDENTIAL_PROVIDER_CREDENTIAL_SERIALIZATION* pcpcsOut)
{
	if (pcpcsIn != NULL && _settings->LoginRemote())
	{
		pcpcsOut = (CREDENTIAL_PROVIDER_CREDENTIAL_SERIALIZATION*)CoTaskMemAlloc(sizeof(CREDENTIAL_PROVIDER_CREDENTIAL_SERIALIZATION));
		pcpcsOut->ulAuthenticationPackage = pcpcsIn->ulAuthenticationPackage;
		pcpcsOut->clsidCredentialProvider = CLSID_PINsafeLoginProvider;
		pcpcsOut->cbSerialization = pcpcsIn->cbSerialization;
		pcpcsOut->rgbSerialization = (LPBYTE)CoTaskMemAlloc(pcpcsOut->cbSerialization);
		memcpy(pcpcsOut->rgbSerialization, pcpcsIn->rgbSerialization, pcpcsOut->cbSerialization);
		return S_OK;
	}
	LogError(L"UpdateRemoteCredential: input credential was NULL");
	return E_NOTIMPL;
}

HRESULT PINsafeLoginFilter_CreateInstance(REFIID riid, void** ppv)
{
    HRESULT hr;

    PINsafeLoginFilter* pFilter = new PINsafeLoginFilter();

    if (pFilter)
    {
        hr = pFilter->QueryInterface(riid, ppv);
        pFilter->Release();
    }
    else
    {
        hr = E_OUTOFMEMORY;
    }
    
    return hr;
}