#include "stdafx.h"
#include "RegSettings.h"

LPCWSTR WINLOGON_KEY = L"SOFTWARE\\Swivel Secure\\Swivel Credential Provider";
LPCWSTR PINSAFE_SERVER_KEY = L"PINsafeServer";
LPCWSTR PINSAFE_PORT_KEY = L"PINsafePort";
LPCWSTR PINSAFE_CONTEXT_KEY = L"PINsafeContext";
LPCWSTR PINSAFE_SECRET_KEY = L"PINsafeSecret";
LPCWSTR PINSAFE_PROTOCOL_KEY = L"PINsafeProtocol";
LPCWSTR PINSAFE_LOGIN_SELECT_KEY = L"PINsafeLoginSelect";
LPCWSTR PINSAFE_SHOW_TURING_KEY = L"PINsafeShowTURing";
LPCWSTR PINSAFE_SHOW_REQUEST_KEY = L"PINsafeRequestString";
LPCWSTR PINSAFE_ALLOW_MSGINA_KEY = L"PINsafeAllowDefaultLogin";
LPCWSTR PINSAFE_IGNORE_DOMAIN_KEY = L"PINsafeIgnoreDomain";
LPCWSTR PINSAFE_ALLOW_UNKNOWN_USERS_KEY = L"PINsafeAllowUnknownUsers";
LPCWSTR PINSAFE_LOGIN_ALWAYS = L"always";
LPCWSTR PINSAFE_LOGIN_REMOTE = L"remote";
LPCWSTR PINSAFE_LOGIN_DISABLED = L"disabled";
LPCWSTR PINSAFE_ALLOW_SELF_CERT_KEY = L"PINsafeAllowSelfCert";
LPCWSTR PINSAFE_DISABLE_FILTER_KEY = L"PINsafeDisableFilter";
LPCWSTR PINSAFE_CACHE_SIZE = L"PINsafeCacheSize";
LPCWSTR PINSAFE_SESSION_TIMEOUT = L"PINsafeSessionTimeout";
LPCWSTR PINSAFE_CACHE_LOW_COUNT = L"PINsafeCacheLowCount";
LPCWSTR PINSAFE_FONTS_FOLDER = L"PINsafeFontsFolder";
LPCWSTR PINSAFE_BACKGROUNDS_FOLDER = L"PINsafeBackgroundsFolder";
LPCWSTR DEFAULT_FONTS_FOLDER = L"fonts";
LPCWSTR DEFAULT_BACKGROUNDS_FOLDER = L"backgrounds";
LPCWSTR PINSAFE_RESOURCE_DLL = L"PINsafeResourceDll";
LPCWSTR DEFAULT_RESOURCE_DLL = L"PINsafeResource.dll";
LPCWSTR PINSAFE_LOCKOUT_COUNT = L"PINsafeLockoutCount";
LPCWSTR PINSAFE_USE_LOCAL = L"PINsafeUseLocalAuth";
LPCWSTR PINSAFE_LOCAL_ALWAYS = L"always";
LPCWSTR PINSAFE_LOCAL_FALLBACK = L"fallback";
LPCWSTR PINSAFE_LOCAL_NEVER = L"never";

static RegSettings* instance = NULL;

// Create the singleton instance;
RegSettings* RegSettings::GetSettings() {
	if (instance == NULL) {
		instance = new RegSettings();
	}
	return instance;
}

//Get the PINsafe Settings from the Registry...
RegSettings::RegSettings(void)
{
	const int bufSize = 0x100;
	wchar_t tempStr[bufSize];

	HKEY hReg;

	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, WINLOGON_KEY, 0, KEY_READ, &hReg) == 0) {

		DWORD s = bufSize;
		DWORD dwType;
		if ((RegQueryValueEx(hReg, PINSAFE_SERVER_KEY, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			server = new wchar_t[s+1];
			wcscpy(server, tempStr);
		} else {
			server = new wchar_t[1];
			*server = 0;
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_PORT_KEY, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			port = new wchar_t[s+1];
			wcscpy(port, tempStr);
		} else {
			port = new wchar_t[5];
			wcscpy(port, L"8080");
		}

		//0 - http 1 - https
		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_PROTOCOL_KEY, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			secure = (wcscmp(L"1", tempStr) == 0) ? TRUE : FALSE;
		} else {
			secure = FALSE;
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_CONTEXT_KEY, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			context = new wchar_t[s+1];
			wcscpy(context, tempStr);
		} else {
			context = new wchar_t[1];
			*context = 0;
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_SECRET_KEY, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			secret = new wchar_t[s+1];
			wcscpy(secret, tempStr);
		} else {
			secret = new wchar_t[1];
			*secret = 0;
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_LOGIN_SELECT_KEY, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			if (_wcsicmp(tempStr, PINSAFE_LOGIN_ALWAYS) == 0) {
				selectLogin = PINsafeLoginAlways;
			} else if (_wcsicmp(tempStr, PINSAFE_LOGIN_REMOTE) == 0) {
				selectLogin = PINsafeLoginRemote;
			} else {
				selectLogin = PINsafeLoginDisabled;
			}
		} else {
			selectLogin = PINsafeLoginDisabled;
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_SHOW_TURING_KEY, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			showGetImage = (wcscmp(L"0", tempStr) ? TRUE : FALSE);
		} else {
			showGetImage = TRUE;
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_SHOW_REQUEST_KEY, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			showRequestString = (wcscmp(L"1", tempStr) ? FALSE : TRUE);
		} else {
			showRequestString = FALSE;
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_IGNORE_DOMAIN_KEY, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			ignoreDomain = (wcscmp(L"1", tempStr) ? FALSE : TRUE);
		} else {
			ignoreDomain = TRUE;
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_ALLOW_UNKNOWN_USERS_KEY, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			allowUnknownUsers = (wcscmp(L"1", tempStr) ? FALSE : TRUE);
		} else {
			allowUnknownUsers = FALSE;
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_DISABLE_FILTER_KEY, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			disableFilter = (wcscmp(L"1", tempStr) ? FALSE : TRUE);
		} else {
			disableFilter = FALSE;
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_ALLOW_MSGINA_KEY, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			allowDefaultLogin = (wcscmp(L"1", tempStr) ? FALSE : TRUE);
		} else {
			allowDefaultLogin = FALSE;
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_ALLOW_SELF_CERT_KEY, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			allowSelfCert = (wcscmp(L"0", tempStr) ? TRUE : FALSE);
		} else {
			allowSelfCert = TRUE;
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_CACHE_SIZE, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			cachedStringCount = _ttoi(tempStr);
			if (cachedStringCount == 0) {
				cachedStringCount = 10;
			}
		} else {
			cachedStringCount = 10;
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_SESSION_TIMEOUT, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			sessionTimeout = _ttoi(tempStr);
			if (sessionTimeout == 0) {
				sessionTimeout = 120000;
			} else {
				sessionTimeout *= 1000;
			}
		} else {
			sessionTimeout = 120000;
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_CACHE_LOW_COUNT, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			cacheLowCount = _ttoi(tempStr);
			if (cacheLowCount == 0) {
				cacheLowCount = 5;
			}
		} else {
			cacheLowCount = 5;
		}

		s = bufSize;
		fontsFolder = NULL;
		if (RegQueryValueEx(hReg, PINSAFE_FONTS_FOLDER, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) {
			if (s > 0) {
				fontsFolder = new wchar_t[s+1];
				wcscpy(fontsFolder, tempStr);
			} else {
				fontsFolder = new wchar_t[wcslen(DEFAULT_FONTS_FOLDER)+1];
				wcscpy(fontsFolder, DEFAULT_FONTS_FOLDER);
			}
		}

		s = bufSize;
		backgroundsFolder = NULL;
		if (RegQueryValueEx(hReg, PINSAFE_BACKGROUNDS_FOLDER, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) {
			if (s > 0) {
				backgroundsFolder = new wchar_t[s+1];
				wcscpy(backgroundsFolder, tempStr);
			} else {
				backgroundsFolder = new wchar_t[wcslen(DEFAULT_BACKGROUNDS_FOLDER)+1];
				wcscpy(backgroundsFolder, DEFAULT_BACKGROUNDS_FOLDER);
			}
		}

		s = bufSize;
		resourceDll = NULL;
		if (RegQueryValueEx(hReg, PINSAFE_RESOURCE_DLL, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) {
			if (s > 0) {
				resourceDll = new wchar_t[s+1];
				wcscpy(resourceDll, tempStr);
			} else {
				resourceDll = new wchar_t[wcslen(DEFAULT_RESOURCE_DLL)+1];
				wcscpy(resourceDll, DEFAULT_RESOURCE_DLL);
			}
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_LOCKOUT_COUNT, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			lockoutCount = _ttoi(tempStr);
		} else {
			lockoutCount = 5;
		}

		s = bufSize;
		if ((RegQueryValueEx(hReg, PINSAFE_USE_LOCAL, NULL, &dwType, (LPBYTE)tempStr, &s) == 0) && (dwType == REG_SZ)) {
			if (_wcsicmp(tempStr, PINSAFE_LOCAL_ALWAYS) == 0) {
				useLocal = PINsafeLocalAlways;
			} else if (_wcsicmp(tempStr, PINSAFE_LOCAL_FALLBACK) == 0) {
				useLocal = PINsafeLocalFallback;
			} else {
				useLocal = PINsafeLocalNever;
			}
		} else {
			useLocal = PINsafeLocalNever;
		}

		RegCloseKey(hReg);
	}
}

RegSettings::~RegSettings(void)
{
	if (server)
		delete server;
	if (port)
		delete port;
	if (context)
		delete context;
	if (secret)
		delete secret;
}
