#include "StdAfx.h"
#include <time.h>
#include "Resource.h"
#include "TuringGenerator.h"
#include "CredentialCache.h"
#include "RegSettings.h"
#include "..\SwivelResource\SwivelResource.h"
#include <ocidl.h>
#include <olectl.h>
#include "Logging.h"

using namespace std;

unsigned int GetRandomNumber(int maxValue)
{
	unsigned int num;
	rand_s(&num);
	num %= maxValue;
	return num;
}

CTuringGenerator::CTuringGenerator(void)
{
	m_securityString = NULL;
	LoadBackgrounds();
	LoadFonts();
}

CTuringGenerator::~CTuringGenerator(void)
{
	if (m_securityString != NULL)
	{
		delete m_securityString;
	}
	for (list<LPCTSTR>::iterator iter = backgrounds.begin(); iter != backgrounds.end(); ) {
		LPCTSTR str = *iter;
		iter = backgrounds.erase(iter);
		delete str;
	}
	for (list<LOGFONT*>::iterator iter = fonts.begin(); iter != fonts.end(); ) {
		LOGFONT* font = *iter;
		iter = fonts.erase(iter);
		delete font;
	}
}

LPCTSTR CTuringGenerator::GetSecurityString(LPCTSTR username)
{
	if (m_securityString != NULL) {
		delete m_securityString;
	}
	m_securityString = NULL;
	CredentialCache* creds = CredentialCache::GetCacheForUser(username);
	LPCTSTR newCredentials = creds->GetSecurityString();
	if (newCredentials == NULL || !(*newCredentials)) {
		return NULL;
	}
	size_t credLen = _tcslen(newCredentials)+1;
	m_securityString = new TCHAR[credLen];
#ifdef UNICODE
	wcscpy(m_securityString, newCredentials);
#else
	WideCharToMultiByte(CP_ACP, 0, newCredentials, -1, m_securityString, credLen, NULL, NULL);
#endif

	return m_securityString;
}

HBITMAP LoadJPGImage(LPCTSTR filename)
{
	HANDLE hFile = CreateFile(filename, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
	if (!hFile || hFile == INVALID_HANDLE_VALUE) {
		return (HBITMAP)INVALID_HANDLE_VALUE;
	}
	DWORD fileSize = GetFileSize(hFile, NULL);
	HGLOBAL hMem = GlobalAlloc(GMEM_MOVEABLE, fileSize);
	LPVOID data = GlobalLock(hMem);
	DWORD bytesRead = 0;
	ReadFile(hFile, data, fileSize, &bytesRead, NULL);
	GlobalUnlock(hMem);
	CloseHandle(hFile);
	LPSTREAM pStrm = NULL;
	CreateStreamOnHGlobal(hMem, TRUE, &pStrm);
	IPicture* pict;
	OleLoadPicture(pStrm, fileSize, FALSE, IID_IPicture, (LPVOID*)(&pict));
	pStrm->Release();
	HBITMAP hBitmap;
	pict->get_Handle((OLE_HANDLE*)&hBitmap);
	return hBitmap;
}

BOOL CTuringGenerator::RenderTURingImage(HWND hWnd, LPCTSTR username, HBITMAP background, int ctrlID) {
	if (hWnd == NULL || hWnd == INVALID_HANDLE_VALUE) {
		return FALSE;
	}
	if (!GetSecurityString(username)) {
		return FALSE;
	}
	HWND image = GetDlgItem(hWnd, ctrlID);
	if (image == NULL) {
		return FALSE;
	}
	RenderSecurityString(image, background, m_securityString);
	return TRUE;
}

void CTuringGenerator::RenderSecurityString(HWND hWnd, HBITMAP background, LPTSTR secString)
{
//	PAINTSTRUCT ps;
	HDC hdc;

	//Draw the background image...
//	hdc = BeginPaint(hWnd, &ps);
	hdc = GetDC(hWnd);

	// Get the frame image
//	HBITMAP bitmap = LoadBitmap(hRes, MAKEINTRESOURCE(IDB_TURING));
//	if (bitmap == NULL || bitmap == INVALID_HANDLE_VALUE) {
//		bitmap = LoadBitmap(hRes, MAKEINTRESOURCE(IDB_BACKGROUND));
//	}

	HDC memDC = CreateCompatibleDC(hdc);
	SelectObject(memDC, background);

	BitBlt(hdc, 0,0, 300, 80, memDC, 0, 0, SRCCOPY);

	DeleteDC(memDC);

	// Get the insert image
	unsigned int imageIdx = GetRandomNumber(backgrounds.size());
	LPCTSTR imageName;
	if (backgrounds.size() > 0) {
		// Believe it or not, there's no way to get a list item by index.
		list<LPCTSTR>::const_iterator imageIter = backgrounds.begin();
		for (; imageIdx > 0 && imageIter != backgrounds.end(); imageIdx--, imageIter++)
			;
		imageName = *imageIter;
		HBITMAP img = LoadJPGImage(imageName);
		// Insert the background image over the frame.
		memDC = CreateCompatibleDC(hdc);
		SelectObject(memDC, img);
		BitBlt(hdc, 10, 38, 280, 32, memDC, 0, 0, SRCCOPY);
		DeleteDC(memDC);
		DeleteObject(img);
	}

	//choose the font
	unsigned int fontIdx = GetRandomNumber(fonts.size());
	HFONT hFont;
	LOGFONT* m_pLF;
	// Believe it or not, there's no way to get a list item by index.
	list<LOGFONT*>::const_iterator fontIter = fonts.begin();
	for (; fontIdx > 0 && fontIter != fonts.end(); fontIdx--, fontIter++)
		;
	m_pLF = *fontIter;
/*	m_pLF=(LOGFONT*)calloc(1,sizeof(LOGFONT));
	wcsncpy_s(m_pLF->lfFaceName,L"Times New Roman",31);
	m_pLF->lfHeight=36;
	m_pLF->lfWeight=600;
	m_pLF->lfItalic=1;
	m_pLF->lfUnderline=0;*/
	hFont = CreateFontIndirect(m_pLF);

	//select the font in the dc
	HFONT oldFont=NULL;
	if (hFont) 
		oldFont = (HFONT)SelectObject(hdc, hFont);
	else 
		hFont = (HFONT)SelectObject(hdc, GetStockObject(DEFAULT_GUI_FONT));

	//Set text color
	SetTextColor(hdc, RGB(0,0,0));

	int psnX = 15;
	int psnY = 48;
	int randY = 0;
	TCHAR secChar[2];
	secChar[1] = 0;
	srand( (unsigned)time( NULL ) );
	for (int i=0; i<10; i++) {				
		randY = rand() & 15;
		while (randY > 8)
		{
			randY = rand() & 15;
		}
		
		
		//Set text position;
		RECT pos = {psnX, (psnY - randY), 0, 0};	

		//Shuffle along to the next one...
		psnX += 28;

		//draw the text
		SetBkMode(hdc, TRANSPARENT);
		secChar[0] = secString[i];
		DrawText(hdc, secChar, 1, &pos, DT_CALCRECT);
		DrawText(hdc, secChar, 1, &pos, 0);
	}
	
	//cleanup 
	if (oldFont) {
		SelectObject(hdc, oldFont);
	}
	DeleteObject(hFont);
//	free(m_pLF);

//	DeleteObject(bitmap);
	
//	EndPaint(hWnd, &ps);
	ReleaseDC(hWnd, hdc);
}

BOOL CTuringGenerator::CheckCredentials(LPCTSTR username, LPCTSTR password, LPCTSTR otc)
{
	// Some basic sanity checks.
	if (m_securityString == NULL || otc == NULL || !(*otc)) {
		return FALSE;
	}

	CredentialCache* creds = CredentialCache::GetCacheForUser(username);
	return creds->Authenticate(m_securityString, password, otc);
}

void CTuringGenerator::LoadFonts()
{
	RegSettings* settings = RegSettings::GetSettings();
	LPCWSTR fontsFolder = settings->GetFontsFolder();
	DWORD lFontsFile;
	LPTSTR fontsListFile;
	if (fontsFolder == NULL) {
		lFontsFile = 16;
		fontsListFile = new TCHAR[lFontsFile];
		_tcscpy(fontsListFile, _T("fonts\\fonts.lst"));
	} else {
		lFontsFile = wcslen(fontsFolder) + 11;
		fontsListFile = new TCHAR[lFontsFile];
#ifdef UNICODE
		_tcscpy(fontsListFile, fontsFolder);
#else
		WideCharToMultiByte(CP_ACP, 0, fontsFolder, -1, fontsListFile, lFontsFile, NULL, NULL);
#endif
		_tcscat(fontsListFile, _T("\\fonts.lst"));
	}
	FILE* hFile;
	if (_tfopen_s(&hFile, fontsListFile, _T("rt"))) {
		LogError(_T("Unable to open fonts list: %s"), fontsListFile);
	} else if (hFile != NULL) {
		TCHAR line[0x100];
		LPTSTR ptr, next_ptr;
		do {
			if (_fgetts(line, 0x100, hFile) == NULL) {
				break;
			}
			DWORD len = _tcslen(line);
			while (len > 0 && isspace(line[len-1])) {
				line[--len] = 0;
			}
			if (!(*line)) {
				continue;
			}
			// Get the font file name
			ptr = _tcstok_s(line, _T(","), &next_ptr);
			if (ptr == NULL) {
				continue;
			}
			DWORD lFontFile = wcslen(fontsFolder) + _tcslen(ptr) + 2;
			LPTSTR fontFile = new TCHAR[lFontFile];
			_tcscpy(fontFile, fontsFolder);
			_tcscat(fontFile, _T("\\"));
			_tcscat(fontFile, ptr);
			AddFontResourceEx(fontFile, FR_PRIVATE | FR_NOT_ENUM, 0);
			// Get the font name
			ptr = _tcstok_s(NULL, _T(","), &next_ptr);
			if (ptr == NULL) {
				// Can't use this font: no name.
				RemoveFontResource(fontFile);
				continue;
			}
			LOGFONT *fontDef = new LOGFONT();
			_tcscpy_s(fontDef->lfFaceName, LF_FACESIZE, ptr);
			// Bold
			ptr = _tcstok_s(NULL, _T(","), &next_ptr);
			if (ptr == NULL) {
				// No more information: use defaults from now on.
				fontDef->lfWeight = FW_NORMAL;
				fontDef->lfItalic = FALSE;
				fontDef->lfHeight = 24;
			} else {
				if (_tcsicmp(ptr, _T("bold")) == 0) {
					fontDef->lfWeight = FW_BOLD;
				} else {
					fontDef->lfWeight = FW_NORMAL;
				}
				ptr = _tcstok_s(NULL, _T(","), &next_ptr);
				if (ptr == NULL) {
					fontDef->lfItalic = FALSE;
					fontDef->lfHeight = 24;
				} else {
					fontDef->lfItalic = (_tcsicmp(ptr, _T("italic")) == 0) ? TRUE : FALSE;
					ptr = _tcstok_s(NULL, _T(","), &next_ptr);
					if (ptr == NULL || *ptr == 0) {
						fontDef->lfHeight = 24;
					} else {
						int fHeight = _ttoi(ptr);
						if (fHeight > 0) {
							fontDef->lfHeight = fHeight;
						} else {
							fontDef->lfHeight = 24;
						}
					}
				}
			}
			fontDef->lfWidth = 0;
			fontDef->lfEscapement = 0;
			fontDef->lfOrientation = 0;
			fontDef->lfUnderline = 0;
			fontDef->lfStrikeOut = 0;
			fontDef->lfCharSet = DEFAULT_CHARSET;
			fontDef->lfOutPrecision = OUT_DEFAULT_PRECIS;
			fontDef->lfClipPrecision = CLIP_DEFAULT_PRECIS;
			fontDef->lfQuality = DEFAULT_QUALITY;
			fontDef->lfPitchAndFamily = DEFAULT_PITCH | FF_DONTCARE;
			fonts.push_back(fontDef);
		} while (TRUE);
		fclose(hFile);
	}
	if (fonts.size() == 0) {
		// Add the default font.
		LOGFONT* m_pLF;
		m_pLF=(LOGFONT*)calloc(1,sizeof(LOGFONT));
		wcsncpy_s(m_pLF->lfFaceName,L"Times New Roman",31);
		m_pLF->lfHeight=36;
		m_pLF->lfWeight=600;
		m_pLF->lfItalic=1;
		m_pLF->lfUnderline=0;
		fonts.push_back(m_pLF);
	}
}

void CTuringGenerator::LoadBackgrounds()
{
	RegSettings* settings = RegSettings::GetSettings();
	LPCWSTR backgroundsFolder = settings->GetBackgroundsFolder();
	DWORD lBGFile;
	LPTSTR BGFile;
	if (backgroundsFolder == NULL) {
		lBGFile = 30;
		BGFile = new TCHAR[lBGFile];
		_tcscpy(BGFile, _T("backgrounds\\backgrounds.lst"));
	} else {
		lBGFile = wcslen(backgroundsFolder) + 17;
		BGFile = new TCHAR[lBGFile];
#ifdef UNICODE
		_tcscpy(BGFile, backgroundsFolder);
#else
		WideCharToMultiByte(CP_ACP, 0, backgroundsFolder, -1, BGFile, lBGFile, NULL, NULL);
#endif
		_tcscat(BGFile, _T("\\backgrounds.lst"));
	}
	FILE* hFile;
	if (_tfopen_s(&hFile, BGFile, _T("rt"))) {
		LogError(_T("Unable to open backgrounds list: %s"), BGFile);
	} else if (hFile != NULL) {
		TCHAR filename[0x20];
		do {
			if (_fgetts(filename, 0x20, hFile) == NULL) {
				break;
			}
			DWORD fileLen = _tcslen(filename);
			while (fileLen > 0 && isspace(filename[fileLen-1])) {
				filename[--fileLen] = 0;
			}
			if (!(*filename)) {
				// Don't include empty names.
				continue;
			}
			DWORD fLen = wcslen(backgroundsFolder) + fileLen + 2;
			LPTSTR bgName = new TCHAR[fLen];
			_tcscpy(bgName, backgroundsFolder);
			_tcscat(bgName, _T("\\"));
			_tcscat(bgName, filename);
			backgrounds.push_back(bgName);
		} while (TRUE);
		fclose(hFile);
	}
}
