// TuringThread.cpp : implementation file
//

#include "stdafx.h"
#include "PINsafeLogin.h"
#include "TuringThread.h"
#include "resource.h"
#include "Logging.h"
#include "PINsafeLoginCredential.h"

extern TuringThread* _turingThread;
extern BOOL _dialogTerminated;

// TuringThread

IMPLEMENT_DYNCREATE(TuringThread, CWinThread)

TuringThread::TuringThread()
{
	_terminated = FALSE;
	_turingRemote = NULL;
	_displayRemote = FALSE;
	_turingLocal = NULL;
	_displayLocal = FALSE;
	_hParentWindow = NULL;
}

TuringThread::~TuringThread()
{
	// Make sure the dialog is closed
	Terminate();
}

void TuringThread::Terminate()
{
	if (_hTURing_Dialog != NULL && _hTURing_Dialog != INVALID_HANDLE_VALUE) {
		ShowWindow(_hTURing_Dialog, SW_HIDE);
		DestroyWindow(_hTURing_Dialog);
	}
	_displayRemote = FALSE;
	if (_turingRemote != NULL) {
		_turingRemote->Release();
		_turingRemote = NULL;
	}
	_displayLocal = FALSE;
	_hParentWindow = NULL;
	_hTURing_Dialog = NULL;
	_hTURing_Image = NULL;
	_hTURing_Bitmap = NULL;
	_terminated = TRUE;
}

INT_PTR CALLBACK TuringProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
		case WM_INITDIALOG:
			// No initialisation required.
			return TRUE;
		case WM_COMMAND:
			// Button clicked: close the window
			PostQuitMessage(IDOK);
			_dialogTerminated = TRUE;
			return TRUE;
	}
	return FALSE;
}

BOOL TuringThread::PreTranslateMessage(MSG *pMsg) {

	if (_terminated) {
		AfxEndThread(0, TRUE);
		return TRUE;
	}

	if (pMsg->hwnd == _hTURing_Dialog) {
		return (IsWindow(_hTURing_Dialog) && IsDialogMessage(_hTURing_Dialog, pMsg));
	}
	return CWinThread::PreTranslateMessage(pMsg);
}

BOOL TuringThread::InitInstance()
{
	_hTURing_Bitmap = (HBITMAP)INVALID_HANDLE_VALUE;
	_hTURing_Image = (HWND)INVALID_HANDLE_VALUE;
	_hTURing_Dialog = CreateDialog(g_hinst, MAKEINTRESOURCE(IDD_TURING), _hParentWindow, TuringProc);
	if (_hTURing_Dialog != NULL && _hTURing_Dialog != INVALID_HANDLE_VALUE) {
		_hTURing_Image = GetDlgItem(_hTURing_Dialog, IDC_TURING);
		ShowWindow(_hTURing_Dialog, SW_SHOWNORMAL);
		if (_displayRemote && !_dialogTerminated) {
			DisplayImage();
		} else if (_displayLocal && !_dialogTerminated) {
			DisplayLocalImage();
		}
		_dialogTerminated = FALSE;
		return TRUE;
	}
	return FALSE;
}

DWORD TuringThread::Resume()
{
	if (_hTURing_Dialog == NULL || _hTURing_Dialog == INVALID_HANDLE_VALUE)
	{
		return FALSE;
	}
	ShowWindow(_hTURing_Dialog, SW_SHOWNORMAL);
	DWORD resume = ResumeThread();
	if (resume == 0 || resume == 1) {
		return resume;
	} else {
		LogError(L"Dialog resume failed: resume count = %d", resume);
	}
	return resume;
}

DWORD TuringThread::Pause()
{
	DWORD result = SuspendThread();
	if (result != 0xffffffff) {
		_displayRemote = FALSE;
		_displayLocal = FALSE;
		ShowWindow(_hTURing_Dialog, SW_HIDE);
	} else {
		LogError(L"Thread suspend failed");
	}
	return result;
}

int TuringThread::ExitInstance()
{
	if (IsWindow(_hTURing_Dialog)) {
		Terminate();
		DestroyWindow(_hTURing_Dialog);
		_hTURing_Dialog = NULL;
	}
	_terminated = TRUE;
	return CWinThread::ExitInstance();
}

void TuringThread::SetImage(IPicture* turing)
{
	if (_turingRemote != NULL) {
		_turingRemote->Release();
	}
	turing->AddRef();
	_displayRemote = TRUE;
	_turingRemote = turing;
	if (_hTURing_Dialog != NULL && IsWindowVisible(_hTURing_Dialog)) {
		// If the dialog is ready, display immediately.
		DisplayImage();
	}
}

void TuringThread::DisplayImage() {
	// Hide the "Waiting for PINsafe" message.
	HWND hWaiting = GetDlgItem(_hTURing_Dialog, IDC_WAITING);
	ShowWindow(hWaiting, SW_HIDE);
	HBITMAP hTemp = NULL;
	_turingRemote->get_Handle((OLE_HANDLE*)&hTemp);
	_hTURing_Bitmap = (HBITMAP)CopyImage(hTemp, IMAGE_BITMAP, 0, 0, LR_COPYRETURNORG);
	if (_hTURing_Bitmap == NULL || _hTURing_Bitmap == INVALID_HANDLE_VALUE) {
		return;
	}
	if (_hTURing_Image == NULL || _hTURing_Image == INVALID_HANDLE_VALUE) {
		return;
	}
	SendMessage(_hTURing_Image, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)_hTURing_Bitmap);
	UpdateWindow(_hTURing_Dialog);
	_pCredential->ShowTURingImageVisible();
	_turingRemote->Release();
	_turingRemote = NULL;
	_displayRemote = FALSE;
}

void TuringThread::SetLocalImage(CTuringGenerator* turing, LPCTSTR username, HINSTANCE hRes) {
	_turingLocal = turing;
	_localUser = username;
	_localResource = hRes;
	_displayLocal = TRUE;
	if (Visible()) {
		// If the dialog is ready, display immediately.
		DisplayLocalImage();
	}
}

void TuringThread::DisplayLocalImage()
{
	HBITMAP background = LoadBitmap(_localResource, MAKEINTRESOURCE(IDB_TURING));
	_turingLocal->RenderTURingImage(_hTURing_Dialog, _localUser, background, IDC_TURING);
	DeleteObject(background);
	if (!UpdateWindow(_hTURing_Dialog)) {
		LONG err = GetLastError();
	}
	_pCredential->ShowTURingImageVisible();
}

BOOL TuringThread::Visible()
{
	return (_hTURing_Dialog != NULL && IsWindowVisible(_hTURing_Dialog));
}

BEGIN_MESSAGE_MAP(TuringThread, CWinThread)
END_MESSAGE_MAP()


// TuringThread message handlers
