//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) 2006 Microsoft Corporation. All rights reserved.
//
// This file contains some global variables that describe what our
// sample tile looks like. For example, it defines which fields to append
// to the wrapped credential's tile, as well as how to display them.

#pragma once
#include <credentialprovider.h>
#include <ntsecapi.h>
#define SECURITY_WIN32
#include <security.h>
#include <intsafe.h>

#define MAX_ULONG  ((ULONG)(-1))

// The indexes of each of the fields in our credential provider's appended tiles.
enum PINSAFE_FIELD_ID 
{
	PFCPI_LABEL			= 0,
	PFCPI_USERNAME		= 1,
    PFCPI_OLD_OTC		= 2,
    PFCPI_NEW_OTC1		= 3,
    PFCPI_NEW_OTC2		= 4,
	PFCPI_SUBMIT		= 5,
	PFCPI_NUM_FIELDS	= 6,
    PFI_OTC				= 6,
    PFI_TURING_BUTTON   = 7,
	PFI_DEMAND_BUTTON	= 8,
    PFI_NUM_FIELDS      = 9, // Note: if new fields are added, keep NUM_FIELDS last.  This is used as a count of the number of fields.
	PFI_USED_FIELDS		= 3  // Field count for scenarios other than change PIN.
};

// The first value indicates when the tile is displayed (selected, not selected)
// the second indicates things like whether the field is enabled, whether it has key focus, etc.
struct FIELD_STATE_PAIR
{
    CREDENTIAL_PROVIDER_FIELD_STATE cpfs;
    CREDENTIAL_PROVIDER_FIELD_INTERACTIVE_STATE cpfis;
};

// These two arrays are seperate because a credential provider might
// want to set up a credential with various combinations of field state pairs 
// and field descriptors.

// The field state value indicates whether the field is displayed
// in the selected tile, the deselected tile, or both.
// The Field interactive state indicates when 
static const FIELD_STATE_PAIR s_rgFieldStatePairs[] = 
{
	{ CPFS_DISPLAY_IN_DESELECTED_TILE, CPFIS_NONE },
	{ CPFS_DISPLAY_IN_SELECTED_TILE, CPFIS_FOCUSED },
    { CPFS_DISPLAY_IN_SELECTED_TILE, CPFIS_NONE },
    { CPFS_DISPLAY_IN_SELECTED_TILE, CPFIS_NONE },
    { CPFS_DISPLAY_IN_SELECTED_TILE, CPFIS_NONE },
    { CPFS_DISPLAY_IN_SELECTED_TILE, CPFIS_NONE },
    { CPFS_DISPLAY_IN_SELECTED_TILE, CPFIS_NONE },
    { CPFS_DISPLAY_IN_SELECTED_TILE, CPFIS_NONE },
    { CPFS_DISPLAY_IN_SELECTED_TILE, CPFIS_NONE }
};

// Field descriptors for unlock and logon.
// The first field is the index of the field.
// The second is the type of the field.
// The third is the name of the field, NOT the value which will appear in the field.
static const CREDENTIAL_PROVIDER_FIELD_DESCRIPTOR s_rgCredProvFieldDescriptors[] =
{
	{ PFCPI_LABEL, CPFT_LARGE_TEXT, L"LABEL" },
	{ PFCPI_USERNAME, CPFT_EDIT_TEXT, L"Username" },
    { PFCPI_OLD_OTC, CPFT_PASSWORD_TEXT, L"Old OTC" },
	{ PFCPI_NEW_OTC1, CPFT_PASSWORD_TEXT, L"New OTC" },
	{ PFCPI_NEW_OTC2, CPFT_PASSWORD_TEXT, L"Confirm New OTC" },
	{ PFCPI_SUBMIT, CPFT_SUBMIT_BUTTON, L"SUBMIT" },
    { PFI_OTC, CPFT_PASSWORD_TEXT, L"OTC" },
    { PFI_TURING_BUTTON, CPFT_COMMAND_LINK, L"START_SESSION" },
	{ PFI_DEMAND_BUTTON, CPFT_COMMAND_LINK, L"REQUEST_STRING" }
};
