//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) 2006 Microsoft Corporation. All rights reserved.
//
// Helper functions for copying parameters and packaging the buffer
// for GetSerialization.

#pragma once
#include "common.h"
#include <windows.h>
#include <strsafe.h>

#pragma warning(push)
#pragma warning(disable : 4995)
#include <shlwapi.h>
#pragma warning(pop)



//makes a copy of a field descriptor using CoTaskMemAlloc
HRESULT FieldDescriptorCoAllocCopy(
    const CREDENTIAL_PROVIDER_FIELD_DESCRIPTOR& rcpfd,
    CREDENTIAL_PROVIDER_FIELD_DESCRIPTOR** ppcpfd
    );

//makes a copy of a field descriptor on the normal heap
HRESULT FieldDescriptorCopy(
    const CREDENTIAL_PROVIDER_FIELD_DESCRIPTOR& rcpfd,
    CREDENTIAL_PROVIDER_FIELD_DESCRIPTOR* pcpfd
    );

//creates a UNICODE_STRING from a normal string
HRESULT UnicodeStringInitWithString(
    PWSTR pwz, 
    UNICODE_STRING* pus
    );

//packages the credentials into the buffer that the system expects
HRESULT KerbInteractiveLogonPack(
    const KERB_INTERACTIVE_LOGON& rkil,
    BYTE** prgb,
    DWORD* pcb
    );

//packages the credentials into the buffer that the system expects
HRESULT KerbChangePasswordPack(
	const KERB_CHANGEPASSWORD_REQUEST& cpr,
    BYTE** prgb,
    DWORD* pcb
    );

//get the authentication package that will be used for our logon attempt
HRESULT RetrieveNegotiateAuthPackage(
    ULONG * pulAuthPackage
    );

// Determine the domain name for a given user.
// Input: the username to check.
// Output: the domain as follows:
//   If a domain is explicitly specified, use that.
//   Otherwise, check if the user exists as a local account, and if so, use the computer name.
//   Failing that, use the domain name, if there is one.
HRESULT GetDomainNameForUser(LPWSTR username, LPWSTR* domain);

class ExpandableString {
public:
	ExpandableString();
	ExpandableString(LPCWSTR str);
	~ExpandableString();
	void Set(LPCWSTR str);
	void Set(LPCWSTR str, size_t size);
	void Append(LPCWSTR str);
	void Append(LPCWSTR str, size_t size);
	void Clear();
	LPCWSTR String() { return _buffer; }
	void CopyToString(LPWSTR* str);
	size_t Length() { return wcslen(_buffer); }
	size_t MaxLength() { return _currentSize; }
	BOOL IsEmpty() { return (*_buffer == 0); }
	void Truncate(size_t size);
	void TruncateLeft(size_t start);
	void Extract(size_t start, size_t size);
	int Locate(WCHAR c);
	int Compare(ExpandableString &str);
	int Compare(LPCWSTR str);
	void ToUnicodeString(UNICODE_STRING &uStr);

private:
	void Initialize(size_t initSize);

	size_t _currentSize;
	LPWSTR _buffer;
};
