using System;
using System.IO;
using System.Xml;
using Microsoft.Win32;

namespace PINsafeLoginConfig {
    public class LoginSettings {
        public const string WINLOGON_KEY = @"SOFTWARE\Swivel Secure\Swivel Credential Provider";
        public const string PINSAFE_SERVER_VALUE = "PINsafeServer";
        public const string PINSAFE_PORT_VALUE = "PINsafePort";
        public const string PINSAFE_CONTEXT_VALUE = "PINsafeContext";
        public const string PINSAFE_SECRET_VALUE = "PINsafeSecret";
        public const string PINSAFE_PROTOCOL_VALUE = "PINsafeProtocol";
        public const string PINSAFE_LOGIN_SELECT_VALUE = "PINsafeLoginSelect";
        public const string PINSAFE_LOGIN_ALWAYS = "always";
        public const string PINSAFE_LOGIN_REMOTE = "remote";
        public const string PINSAFE_LOGIN_DISABLED = "disabled";
        public const string PINSAFE_SHOW_TURING_VALUE = "PINsafeShowTURing";
        public const string PINSAFE_ALLOW_MSGINA_VALUE = "PINsafeAllowDefaultLogin";
        public const string PINSAFE_ALLOW_SELF_CERT_VALUE = "PINsafeAllowSelfCert";
        public const string PINSAFE_SHOW_REQUEST_STRING = "PINsafeRequestString";
        public const string PINSAFE_DISABLE_FILTER = "PINsafeDisableFilter";
        public const string PINSAFE_USE_LOCAL = "PINsafeUseLocalAuth";
        public const string PINSAFE_LOCAL_ALWAYS = "always";
        public const string PINSAFE_LOCAL_FALLBACK = "fallback";
        public const string PINSAFE_LOCAL_NEVER = "never";
        public const string PINSAFE_IGNORE_DOMAIN = "PINsafeIgnoreDomain";
        public const string PINSAFE_ALLOW_NON_PINSAFE_USERS = "PINsafeAllowUnknownUsers";
        private const string XML_CONFIG_ROOT = "PINsafeLoginConfig";
        private const string XML_CONFIG_NAMESPACE = "http://schema.swivelsecure.com/login";
        private const string XML_CONFIG_SERVER = "Server";
        private const string XML_CONFIG_PORT = "Port";
        private const string XML_CONFIG_CONTEXT = "Context";
        private const string XML_CONFIG_SECRET = "Secret";
        private const string XML_CONFIG_SECURE = "Secure";
        private const string XML_CONFIG_ALLOWSELFCERT = "AllowSelfCert";
        private const string XML_CONFIG_LOGINSELECT = "LoginSelect";
        private const string XML_CONFIG_SHOWTURING = "ShowTuring";
        private const string XML_CONFIG_SHOWREQUESTSTRING = "ShowRequestString";
        private const string XML_CONFIG_TESTMODE = "TestMode";
        private const string XML_CONFIG_ALLOWDEFAULT = "AllowDefault";
        private const string XML_CONFIG_USELOCAL = "UseLocal";
        private const string XML_CONFIG_IGNOREDOMAIN = "IgnoreDomain";
        private const string XML_CONFIG_ALLOWUNKNOWNUSERS = "AllowUnknownUsers";

        protected RegistryKey _key;

        public LoginSettings() {
            try {
                _key = Registry.LocalMachine.OpenSubKey(WINLOGON_KEY, true);
            } catch (Exception) {
                _key = null;
            }
        }

        ~LoginSettings() {
            if (_key != null) {
                _key.Close();
            }
        }

        public void Save() {
            _key.Flush();
        }

        public string Server {
            get { return (string)_key.GetValue(PINSAFE_SERVER_VALUE, ""); }
            set { _key.SetValue(PINSAFE_SERVER_VALUE, value); }
        }

        public string Port {
            get { return (string)_key.GetValue(PINSAFE_PORT_VALUE, "8080"); }
            set { _key.SetValue(PINSAFE_PORT_VALUE, value); }
        }

        public string Context {
            get { return (string)_key.GetValue(PINSAFE_CONTEXT_VALUE, ""); }
            set { _key.SetValue(PINSAFE_CONTEXT_VALUE, value); }
        }

        public string Secret {
            get { return (string)_key.GetValue(PINSAFE_SECRET_VALUE, ""); }
            set { _key.SetValue(PINSAFE_SECRET_VALUE, value); }
        }

        public bool Secure {
            get { return ((string)_key.GetValue(PINSAFE_PROTOCOL_VALUE, "0") == "1"); }
            set { _key.SetValue(PINSAFE_PROTOCOL_VALUE, value ? "1" : "0"); }
        }

        public bool Valid {
            get { return (_key != null); }
        }

        // TODO Perhaps use an enum?
        public string LoginSelect {
            get { return (string)_key.GetValue(PINSAFE_LOGIN_SELECT_VALUE, PINSAFE_LOGIN_DISABLED); }
            set { _key.SetValue(PINSAFE_LOGIN_SELECT_VALUE, value); }
        }

        public bool ShowTuring {
            get { return (string)_key.GetValue(PINSAFE_SHOW_TURING_VALUE, "1") != "0"; }
            set { _key.SetValue(PINSAFE_SHOW_TURING_VALUE, value ? "1" : "0"); }
        }

        public bool ShowRequestString {
            get { return (string)_key.GetValue(PINSAFE_SHOW_REQUEST_STRING, "1") != "0"; }
            set { _key.SetValue(PINSAFE_SHOW_REQUEST_STRING, value ? "1" : "0"); }
        }

        public bool AllowDefault {
            get { return ((string)_key.GetValue(PINSAFE_ALLOW_MSGINA_VALUE, "0") == "1"); }
            set { _key.SetValue(PINSAFE_ALLOW_MSGINA_VALUE, value ? "1" : "0"); }
        }

        public bool AllowSelfCert {
            get { return (string)_key.GetValue(PINSAFE_ALLOW_SELF_CERT_VALUE, "1") != "0"; }
            set { _key.SetValue(PINSAFE_ALLOW_SELF_CERT_VALUE, value ? "1" : "0"); }
        }

        public bool TestMode {
            get { return ((string)_key.GetValue(PINSAFE_DISABLE_FILTER, "0") == "1"); }
            set { _key.SetValue(PINSAFE_DISABLE_FILTER, value ? "1" : "0"); }
        }

        public string UseLocal {
            get { return ((string)_key.GetValue(PINSAFE_USE_LOCAL, PINSAFE_LOCAL_NEVER)); }
            set { _key.SetValue(PINSAFE_USE_LOCAL, value); }
        }

        public bool IgnoreDomain {
            get { return ((string)_key.GetValue(PINSAFE_IGNORE_DOMAIN, "1") == "1"); }
            set { _key.SetValue(PINSAFE_IGNORE_DOMAIN, value ? "1" : "0"); }
        }

        public bool AllowUnknownUsers {
            get { return ((string) _key.GetValue(PINSAFE_ALLOW_NON_PINSAFE_USERS, "0") == "1"); }
            set { _key.SetValue(PINSAFE_ALLOW_NON_PINSAFE_USERS, value ? "1" : "0"); }
        }

        public void SaveToFile(string filename) {
            using (StreamWriter writer = new StreamWriter(filename)) {
                XmlWriter xmlWriter = XmlWriter.Create(writer, new XmlWriterSettings { Indent = true, CloseOutput = false });
                xmlWriter.WriteStartDocument();
                xmlWriter.WriteStartElement(XML_CONFIG_ROOT, XML_CONFIG_NAMESPACE);
                xmlWriter.WriteElementString(XML_CONFIG_SERVER, Server);
                xmlWriter.WriteElementString(XML_CONFIG_PORT, Port);
                xmlWriter.WriteElementString(XML_CONFIG_CONTEXT, Context);
                xmlWriter.WriteElementString(XML_CONFIG_SECRET, Secret);
                xmlWriter.WriteElementString(XML_CONFIG_SECURE, Secure.ToString());
                xmlWriter.WriteElementString(XML_CONFIG_ALLOWSELFCERT, AllowSelfCert.ToString());
                xmlWriter.WriteElementString(XML_CONFIG_LOGINSELECT, LoginSelect);
                xmlWriter.WriteElementString(XML_CONFIG_SHOWTURING, ShowTuring.ToString());
                xmlWriter.WriteElementString(XML_CONFIG_SHOWREQUESTSTRING, ShowRequestString.ToString());
                xmlWriter.WriteElementString(XML_CONFIG_TESTMODE, TestMode.ToString());
                xmlWriter.WriteElementString(XML_CONFIG_ALLOWDEFAULT, AllowDefault.ToString());
                xmlWriter.WriteElementString(XML_CONFIG_USELOCAL, UseLocal);
                xmlWriter.WriteElementString(XML_CONFIG_IGNOREDOMAIN, IgnoreDomain.ToString());
                xmlWriter.WriteElementString(XML_CONFIG_ALLOWUNKNOWNUSERS, AllowUnknownUsers.ToString());
                xmlWriter.WriteEndElement();
                xmlWriter.Flush();
            }
        }

        public void LoadFromFile(string filename) {
            using (StreamReader reader = new StreamReader(filename)) {
                string xmlInput = reader.ReadToEnd();
                XmlDocument settingsDoc = new XmlDocument();
                settingsDoc.LoadXml(xmlInput);
                XmlElement settingsElem = settingsDoc.DocumentElement;
                if (settingsElem == null) {
                    throw new XmlException("Empty Document");
                }
                string ns = settingsElem.NamespaceURI;
                Server = GetElementValue(settingsElem, ns, XML_CONFIG_SERVER, "");
                Port = GetElementValue(settingsElem, ns, XML_CONFIG_PORT, "8080");
                Context = GetElementValue(settingsElem, ns, XML_CONFIG_CONTEXT, "pinsafe");
                Secret = GetElementValue(settingsElem, ns, XML_CONFIG_SECRET, "");
                Secure = GetElementBool(settingsElem, ns, XML_CONFIG_SECURE, true);
                AllowSelfCert = GetElementBool(settingsElem, ns, XML_CONFIG_ALLOWSELFCERT, true);
                LoginSelect = GetElementValue(settingsElem, ns, XML_CONFIG_LOGINSELECT, PINSAFE_LOGIN_DISABLED);
                ShowTuring = GetElementBool(settingsElem, ns, XML_CONFIG_SHOWTURING, true);
                ShowRequestString = GetElementBool(settingsElem, ns, XML_CONFIG_SHOWREQUESTSTRING, false);
                TestMode = GetElementBool(settingsElem, ns, XML_CONFIG_TESTMODE, false);
                AllowDefault = GetElementBool(settingsElem, ns, XML_CONFIG_ALLOWDEFAULT, false);
                UseLocal = GetElementValue(settingsElem, ns, XML_CONFIG_USELOCAL, PINSAFE_LOCAL_NEVER);
                IgnoreDomain = GetElementBool(settingsElem, ns, XML_CONFIG_IGNOREDOMAIN, true);
                AllowUnknownUsers = GetElementBool(settingsElem, ns, XML_CONFIG_ALLOWUNKNOWNUSERS, false);
            }
        }

        private static bool GetElementBool(XmlElement settingsElem, string ns, string elemName, bool defaultBool) {
            string sBoolean = GetElementValue(settingsElem, ns, elemName, defaultBool.ToString());
            bool boolVal;
            if (!Boolean.TryParse(sBoolean, out boolVal)) {
                boolVal = defaultBool;
            }
            return boolVal;
        }

        private static string GetElementValue(XmlElement settingsElem, string ns, string elemName, string defaultValue) {
            XmlElement valueElem = settingsElem[elemName, ns];
            return (valueElem == null) ? defaultValue : valueElem.InnerText;
        }
    }
}
