using System;
using System.IO;
using System.Windows.Forms;

namespace PINsafeLoginConfig
{
    static class Program
    {
        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static void Main(string[] args)
        {
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);

			try
			{
				LoginSettings settings = new LoginSettings();
				if (settings.Valid)
				{
					if (args.Length > 0)
					{
						string filename = args[0];
						if (File.Exists(filename))
						{
							try
							{
								settings.LoadFromFile(filename);
							}
							catch (Exception)
							{
								MessageBox.Show(String.Format("The settings file \"{0}\" does not appear to be valid", filename),
								                "PINsafe Login Configuration", MessageBoxButtons.OK, MessageBoxIcon.Error);
							}
						}
						else
						{
							MessageBox.Show(String.Format("The settings file \"{0}\" cannot be found", filename),
											"PINsafe Login Configuration", MessageBoxButtons.OK, MessageBoxIcon.Error);
						}
					}
					else
					{
						Application.Run(new SettingsForm(settings));
					}
				}
				else
				{
					MessageBox.Show("You do not have sufficient permissions to run this application", "PINsafe Login Configuration",
								MessageBoxButtons.OK, MessageBoxIcon.Error);
				}

			}
			catch (Exception e)
			{
                MessageBox.Show(String.Format("An error occurred trying to run PINsafe Login Configuration:\n{0}\n{1}", e.Message, e.StackTrace), "PINsafe Login Configuration",
				                MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
		}
    }
}